/*
 * Main routine for the Calvin chatserver.
 *
 * This module handles the main "wait for new connections or new data from old
 * connections, and hand off handling them to others" loop, as well as initial
 * setup.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 1995 Jonathan Lennox
 * Copyright 2017 Russ Allbery <eagle@eyrie.org>
 *
 * See LICENSE for licensing terms.
 */

#include <config.h>
#include <portable/system.h>

#include <errno.h>

#include <util/macros.h>

#include "channels.h"
#include "connections.h"
#include "config-file.h"
#include "format.h"
#include "nicks.h"
#include "init.h"

char *progname;
int dns_fd; /* This needs to be global unfortunately */

static void greet_user(struct connection *newcon) {
  write_to_user(formatn("motd", newcon, NULL), newcon);
  write_to_user(formatn("nick_request", newcon, NULL), newcon);
}

int main(int argc UNUSED, char *argv[])
{
  int retval;
  int incoming_fd, ssl_incoming_fd;
  struct connection *current, *next;

  progname = argv[0];

  initialize_server(&incoming_fd, &ssl_incoming_fd, &dns_fd);
  ssl_init();

  while (1) {
    fd_set fds, wfds;

    /* XXX optimize this so we're not traversing the list every time
       through? */
    FD_ZERO(&fds);
    FD_ZERO(&wfds);
    FD_SET(incoming_fd, &fds);
#ifdef DO_SSL
    FD_SET(ssl_incoming_fd, &fds);
#endif
#ifndef __CYGWIN32__
    FD_SET(dns_fd, &fds);
#endif
    for (current = connections; current != NULL; current = current->next) {
      FD_SET(current->fd, &fds);
      if(current->waiting) FD_SET(current->fd, &wfds);
    }

    retval = select(FD_SETSIZE, &fds, &wfds, NULL, NULL);
    if (retval < 0)
      fprintf(stderr, "%s: error while selecting: %s\n",
              progname, strerror(errno));
    else if (retval == 0)
      fprintf(stderr, "%s: confused while selecting\n", progname);
    else {
#ifdef DO_SSL
      /* Accept SSL connections. */
      if (FD_ISSET(ssl_incoming_fd, &fds)) {
        struct connection * newcon;

        newcon = setup_new_ssl_connection(ssl_incoming_fd);
        if (newcon == NULL)
          fprintf(stderr, "%s: error setting up new connection: %s\n",
                  progname, strerror(errno));
        else {
          if(newcon->ssl_finished) greet_user(newcon);
        }
      }
#endif
      /* Accept connections. */
      if (FD_ISSET(incoming_fd, &fds)) {
        struct connection * newcon;

        newcon = setup_new_connection(incoming_fd);
        if (newcon == NULL)
          fprintf(stderr, "%s: error setting up new connection: %s\n",
                  progname, strerror(errno));
        else {
          greet_user(newcon);
        }
      }
      /* Accept DNS info. */
      if (FD_ISSET(dns_fd, &fds)) {
        do_dns_received(dns_fd);
      }
      /* Sockets available for writing? */
      for (current = connections; current != NULL; current = next) {
        next = current->next;
        if(FD_ISSET(current->fd, &wfds)) {
          attempt_write(current);
        }
      }
      /* Sockets available for reading? */
      for (current = connections; current != NULL; current = next) {
        /* Note that "current" might become invalidated by functions
           called under read_from_user. */
        next = current->next;
        if (FD_ISSET(current->fd, &fds)) {
#if DO_SSL
          if ((current->ssl) && (!(current->ssl_finished))) {
            continue_ssl_setup(current);
            if(current->ssl_finished) greet_user(current);
          } else
#endif
          {
            retval = read_from_user(current);
            if (retval > 0)
              fprintf(stderr, "%s: connection %d (%s from %s): error "
                      "reading from connection: %s\n",
                      progname, current->fd, nickname(current),
                      current->hostname, strerror(retval));
            else if (retval < 0 && retval != EOF)
              fprintf(stderr, "%s: connection %d (%s from %s): error %d "
                      "reading from connection.\n",
                      progname, current->fd, nickname(current),
                      current->hostname, retval);
            /* EOF is normal and expected. */
          }
        }
      }
      /* Done with select(). */
      while (reap) {
        /* Clean up dead connections. */
        /* This has to be done here rather than at the time we discover the
           death, or the linked list could get mangled. */
        reap = 0;
        for (current = connections; current != NULL; current = next) {
          /* Note that "current" becomes invalid after it's reaped. */
          next = current->next;
          if (current->reap_me) {
            close(current->fd);
#ifdef DO_SSL
            if(current->ssl) ssl_end(current->ssl);
#endif
            close_all_channels(current);
            if (connections == current) connections = current->next;
            if (current->next) current->next->prev = current->prev;
            if (current->prev) current->prev->next = current->next;
            /* If we buffered an announcement for an error disconnect,
               handle it here. */
            if(current->error) {
              announce_disconnection(current, current->error);
              free(current->error);
            }
            free (current->hostname);
#if DO_UNORTHODOX
            if (current->unorthodox) {
              int i;
              if(current->unorthoname)
                for(i = 0; i < current->unorthocount; i++)
                  free(current->unorthoname[i]);
              free(current->unorthoname);
            }
#endif
            free (current->nick);
            free (current);
          }
        }
      }
    }
  }
  return 0;
}
