/*
 * Calvin chatserver command implementation.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 1995 Jonathan Lennox
 * Copyright 2017 Russ Allbery <eagle@eyrie.org>
 *
 * See LICENSE for licensing terms.
 */

/* AIX requires this to be the first thing in the file... */
#if !defined (__GNUC__) && defined (_AIX)
 #pragma alloca
#endif

#include <config.h>
#include <portable/system.h>

#include <ctype.h>
#include <errno.h>

#ifdef __GNUC__
# undef alloca
# define alloca(size) __builtin_alloca(size)
#else
# if HAVE_ALLOCA_H
#  include <alloca.h>
# else
#  if !defined(_AIX) && !defined(alloca) /* predefined by HP cc +Olibcalls */
char *alloca ();
#  endif
# endif
#endif

#include <util/macros.h>

#include "channels.h"
#include "connections.h"
#include "config-file.h"
#include "commands.h"
#include "format.h"
#include "nicks.h"
#include "lastlog.h"
#include "misc.h"

#define min(a,b) (((a)<(b))?(a):(b))

struct command {
  const char *name;
  const char *desc;
  const char *longdesc;
  enum priv privs_required;
  int describe;
  int (*command)(const char *cmd_name, struct connection *user, char *args);
};

int help_cmd(const char *cmd_name, struct connection *user, char *args);
int privilege_cmd(const char *cmd_name, struct connection *user, char *args);

static struct command commands[] =
{
  { "me",       "pose_desc",     "pose_longdesc",     0, 1, pose_cmd },
  { "do",       "pose_desc",     "pose_longdesc",     0, 0, pose_cmd },
  { "pose",     "pose_desc",     "pose_longdesc",     0, 0, pose_cmd },
  { "my",       "ppose_desc",    "ppose_longdesc",    0, 1, ppose_cmd },
  { "ppose",    "ppose_desc",    "ppose_longdesc",    0, 0, ppose_cmd },
  { "nickname", "nickname_desc", "nickname_longdesc",-1, 1, nickname_cmd },
  { "channel",  "channel_desc",  "channel_longdesc",  0, 1, channel_cmd },
  { "join",     "join_desc",     "join_longdesc",     0, 1, join_cmd },
  { "leave",    "leave_desc",    "leave_longdesc",    0, 1, leave_cmd },
  { "part",     "leave_desc",    "leave_longdesc",    0, 0, leave_cmd },
  { "hide",     "hide_desc",     "hide_longdesc",     0, 0, hide_cmd },
  { "topic",    "topic_desc",    "topic_longdesc",    0, 1, topic_cmd },
  { "list",     "list_desc",     "list_longdesc",    -1, 1, list_cmd },
  { "who",      "list_desc",     "list_longdesc",    -1, 0, list_cmd },
  { "listall",  "listall_desc",  "listall_longdesc", -1, 1, listall_cmd },
  { "la",       "listall_desc",  "listall_longdesc", -1, 0, listall_cmd },
  { "whoall",   "listall_desc",  "listall_longdesc", -1, 0, listall_cmd },
  { "users",    "users_desc",    "users_longdesc",   -1, 1, users_cmd },
  { "whois",    "whois_desc",    "whois_longdesc",   -1, 1, whois_cmd },
  { "stat",     "whois_desc",    "whois_longdesc",   -1, 0, whois_cmd },
  { "wi",       "whois_desc",    "whois_longdesc",   -1, 0, whois_cmd },
  { "lastlog",  "lastlog_desc",  "lastlog_longdesc", -1, 1, lastlog_cmd },
  { "log",      "lastlog_desc",  "lastlog_longdesc", -1, 0, lastlog_cmd },
  { "whisper",  "whisper_desc",  "whisper_longdesc",  0, 1, whisper_cmd },
  { "msg",      "whisper_desc",  "whisper_longdesc",  0, 0, whisper_cmd },
  { "message",  "whisper_desc",  "whisper_longdesc",  0, 0, whisper_cmd },
  { "privilege","privilege_desc","privilege_longdesc",0, 1, privilege_cmd },
  { "kick",     "kick_desc",     "kick_longdesc",     2, 1, kick_cmd },
  { "reload",   "reload_desc",   "reload_longdesc",   2, 1, reload_cmd },
  { "wrap",     "wrap_desc",     "wrap_longdesc",    -1, 1, wrap_cmd },
  { "client",   "client_desc",   "client_longdesc",  -1, 0, client_cmd },
  { "date",     "date_desc",     "date_longdesc",    -1, 1, date_cmd },
  { "time",     "date_desc",     "date_longdesc",    -1, 0, date_cmd },
  { "clock",    "date_desc",     "date_longdesc",    -1, 0, date_cmd },
  { "motd",     "motd_desc",     "motd_longdesc",    -1, 1, motd_cmd },
  { "help",     "help_desc",     "help_longdesc",    -1, 1, help_cmd },
  { "quit",     "quit_desc",     "quit_longdesc",    -1, 1, bye_cmd },
  { "bye",      "quit_desc",     "quit_longdesc",    -1, 0, bye_cmd },
  { "off",      "quit_desc",     "quit_longdesc",    -1, 0, bye_cmd },
  { "logoff",   "quit_desc",     "quit_longdesc",    -1, 0, bye_cmd },
  { "signoff",  "quit_desc",     "quit_longdesc",    -1, 0, bye_cmd },
  { "exit",     "quit_desc",     "quit_longdesc",    -1, 0, bye_cmd },
#if DO_YAY
  { "yay",      "yay_desc",      "yay_longdesc",      0, 0, yay_cmd },
#endif
#if DO_ROLL
  { "forgeroll","forgeroll_desc","forgeroll_longdesc",1, 1, forgeroll_cmd },
  { "roll",     "roll_desc",     "roll_longdesc",     0, 1, roll_cmd },
  { "croll",    "croll_desc",    "croll_longdesc",    0, 0, roll_cmd },
#endif
#if DO_YELL
  { "yell",     "yell_desc",     "yell_longdesc",     0, 1, yell_cmd },
#endif
#if DO_UNORTHODOX
  { "trudy",    "trudy_desc",    "trudy_longdesc",    0, 0, trudy_cmd },
  { "notrudy",  "notrudy_desc",  "notrudy_longdesc",  0, 0, notrudy_cmd },
#endif
#if DO_CLOAK
  { "cloak",    "cloak_desc",    "cloak_longdesc",    2, 0, cloak_cmd },
  { "nocloak",  "uncloak_desc",  "uncloak_longdesc",  2, 0, uncloak_cmd },
#endif
#if DO_ALIASES
  { "as",       "as_desc",       "as_longdesc",       0, 1, as_cmd },
  { "narrate",  "narrate_desc",  "narrate_longdesc",  0, 1, narrate_cmd },
#endif
  { NULL,       NULL,            NULL,                0, 0, NULL }
};

static char *upcase(char *string)
{
  int i;

  for (i = 0; string[i] != '\0'; i++)
    if (islower(string[i]))
      string[i] -= ('a' - 'A');

  return string;
}

/* Find the command matching the string.  -1 means not found, -2 means
   ambiguous. */
static int lookup_command(char *command)
{
  int i, found = -1;
  size_t command_len;

  command_len = strlen(command);

  for (i = 0; commands[i].name != NULL; i++) {
    if (command_len <= strlen(commands[i].name) &&
        strncasecmp(command, commands[i].name, command_len) == 0) {
      /* command_to_exec is a substring of commands[i].name */
      /* If they match exactly, take it, otherwise it might be
         ambiguous. */
      if (strcasecmp(command, commands[i].name) == 0) {
        found = i;
        break;
      }
      if (found == -1) {
        found = i;
      } else {
        /* If one of the commands is a substring of the other, go with
           the shorter one, otherwise we're ambiguous. */
        int cmd_i_len = strlen(commands[i].name);
        int cmd_found_len = strlen(commands[found].name);

        if (strncasecmp(commands[i].name, commands[found].name,
                        min(cmd_i_len, cmd_found_len)) == 0) {
          if (cmd_i_len < cmd_found_len)
            found = i;
          /* else leave found alone */
        } else {
          /* We're ambiguous */
          found = -2;
          break;
        }
      }
    }
  }
  return found;
}

/* Parse and dispatch commands. -errno on error. */
int exec_command(char *line, struct connection *user)
{
  int i, ch;
  int found = -1;
  char *command_to_exec, *args, *ptr;
  int command_len;

  for (i = 0; line[i] != '\0' && !isspace(line[i]); i++)
    ; /* Do nothing */

  command_len = i;

  command_to_exec = (char *) alloca(command_len+1);
  if (command_to_exec == NULL) return -ENOMEM;
  strncpy(command_to_exec, line, command_len);
  command_to_exec[command_len] = '\0';

  for (; line[i] != '\0' && isspace(line[i]); i++)
    ; /* Do nothing */

  args = &(line[i]);

  /* Special cases. */
  ch = strtol(command_to_exec, &ptr, 0);
  if (command_to_exec != ptr && *ptr == '\0' && valid_channel(ch)) {
    /* @0 - @15 */
    if (user->privs >= normal_priv) { /* This should match the privs of join
                                         and whisper, above. */
      if (args[0] == '\0')
        /* Change to the specified channel */
        return join_cmd(command_to_exec, user, command_to_exec);
      else
        /* Send to the specified channel */
        return whisper_cmd(command_to_exec, user, line);
    } else {
      return write_to_user(formatn("insufficient_privs",
                                  user, command_to_exec),
                           user);
    }
  } else {
    found = lookup_command(command_to_exec);

    upcase(command_to_exec);

    if (found == -1) {
#ifdef DO_HELPERS
      char *val, *executor;
      struct connection *target;

      val = alloca(strlen(command_to_exec) + 5);
      sprintf(val, "cmd_%s", command_to_exec);
      executor = lookup_conf_no_error(val);
      if(executor) {
        char chanstr[16];
        if(user->target.argtype == RCPT_CHANNEL)
          sprintf(chanstr, "%d", user->target.u.channel);
        target = lookup_nickname(executor);

        if(target && (target->privs >= srv_priv)) {
          if(write_to_user(formatv("$$$ USER $n $p $c [$C] $sa $sb $sc\n", user, 7001, 3,
                                   user->target.argtype == RCPT_UNSPEC ? "0" :
                                   user->target.argtype == RCPT_USER ? "1" :
                                   user->target.argtype == RCPT_CHANNEL ? "2" :
                                   "3" /* RCPT_EVERYONE */,
                                   user->target.argtype == RCPT_USER ?
                                     nickname(user->target.u.user) :
                                   user->target.argtype == RCPT_CHANNEL ?
                                     chanstr : "0",
                                   args),
                          target)) {
            return write_to_user(formatn("failed_executor", user,
                                 command_to_exec), user);
          } else {
            return 0;
          }
        } else {
          return write_to_user(formatn("failed_executor", user,
                               command_to_exec), user);
        }
      } else {
        return write_to_user(formatn("unknown_command",
                                    user, command_to_exec),
                             user);
      }
#else
      return write_to_user(formatn("unknown_command",
                                  user, command_to_exec),
                           user);
#endif
    }
    else if (found == -2)
      return write_to_user(formatn("ambiguous_command",
                                  user, command_to_exec),
                           user);
    else if (found < 0) {
      /* This shouldn't happen */
      return write_to_user(format(NULL, user, command_to_exec),
                           user);
    }
    else if (commands[found].privs_required > user->privs) {
      return write_to_user(formatn("insufficient_privs",
                                  user, command_to_exec),
                           user);
    }
    else
      return (*(commands[found].command))(command_to_exec, user, args);
   }
}


/* Commands */
int help_cmd(const char *cmd_name UNUSED, struct connection *user, char *args)
{
  const char *desc;
  char *fullname;
  int i, retval;

  if (strcmp(args, "") == 0) {
    retval = write_to_user(formatn("help_text", user, NULL), user);
    if (retval) return retval;

    for (i = 0; commands[i].name != NULL; i++) {
      if (commands[i].describe == 1 &&
          commands[i].privs_required <= user->privs) {
        desc = lookup_conf(commands[i].desc);
        retval = write_to_user(format2n("help_format", user,
                               commands[i].name, desc), user);
        if (retval) return retval;
      }
    }
    return 0;
  } else {
    int cmd;
    while (args[0] == '/' || args[0] == '@')
      args++;

    cmd = lookup_command(args);

    upcase(args);

    if (cmd == -1)
      return write_to_user(formatn("unknown_command",
                                  user, args),
                           user);
    else if (cmd == -2)
      return write_to_user(formatn("ambiguous_command",
                                  user, args),
                           user);
    else if (cmd < 0) {
      /* This shouldn't happen */
      return write_to_user(format(NULL, user, args), user);
    }

    /* XXX do I want this here? */
    if (commands[cmd].privs_required > user->privs) {
      return write_to_user(formatn("insufficient_privs",
                                  user, args),
                           user);
    }
    fullname = upcase(strdup(commands[cmd].name));
    retval = write_to_user(formatn(commands[cmd].longdesc,
                                  user, fullname),
                           user);
    free(fullname);
    return retval;
  }
}

int privilege_cmd(const char *cmd_name UNUSED, struct connection *user,
                  char *args)
{
  /* This can only be executed at level 0 or above... */
  int newlevel, retval;
  char *ptr, pr[3], pw_lookup[12];
  const char *enc_pw;

  if (*args == '\0') {
    return write_to_user(formatn("current_privs",
                                user, NULL),
                         user);
  }

  newlevel = strtol(args, &ptr, 0);

  if (ptr != args) {
    /* If they've handed us something bizarre, just fail silently. */

    if (newlevel >= LAST_PRIV || newlevel < normal_priv) {
      *ptr = '\0';
      return write_to_user(formatn("invalid_priv_level",
                                  user, args),
                           user);
    }
    sprintf(pr, "%d", newlevel);

    if (newlevel > user->privs) {
      /* We need a password */
      while (*ptr != '\0' && isspace(*ptr))
        ptr++;

      sprintf(pw_lookup, "priv_%d_pass", newlevel);
      enc_pw = lookup_conf(pw_lookup);
      if (enc_pw == NULL)
        /* This shouldn't happen -- if it does, the config file is set up
           wrong, and we'll have a warning in the error log from that. */
        return 0;
      retval = strcmp(crypt(ptr, enc_pw /* salt */), enc_pw);
      memset(ptr, 0, strlen(ptr));
      if (retval != 0)
        return write_to_user(formatn("invalid_password",
                                    user, pr),
                             user);
    }
    user->privs = newlevel;
    return write_to_user(formatn("new_priv_level",
                                user, NULL),
                         user);
  }
  return 0;
}
