/*
 * Read the config file and give back variables, etc.
 *
 * Syntax of the config file:
 *
 *     variablename "value"
 *     variablename "value"
 *     ...
 *
 *     # to EOL is a comment, except within quotes
 *
 * All C \-escapes are supported within quotes.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 1995 Jonathan Lennox
 * Copyright 2017 Russ Allbery <eagle@eyrie.org>
 *
 * See LICENSE for licensing terms.
 */

#include <config.h>
#include <portable/system.h>

#include <errno.h>
#include <ctype.h>
#include <regex.h>

#include <util/macros.h>

#include "hsearch.h"

#include "channels.h"
#include "connections.h"
#include "format.h"
#include "dyn-str.h"
#include "scan-config.h"
#include "config-file.h"
#include "paths.h"
#include "numerics.h"

struct value {
  char *value;
  int numeric;
  int value_len;
};

/* Now defined in scan-config.h
struct configuration {
  char *data;
  char *data_ptr;
  char *data_end;
  size_t data_size;

  struct value *values;
  struct value *value_ptr;
  struct value *values_end;
  size_t value_table_size;

  struct hsearch_data *htab;

  regex_t regexps[REGEX_COUNT];

  char client_chars[5];
};
*/

extern char *progname;

struct configuration *config = NULL;

static int init_config(struct configuration **confp);
static int setup_regexps(struct configuration *conf);
static int setup_client_chars(struct configuration *conf);
static char *setup_ban_regexp(struct configuration *conf);
static void free_config(struct configuration *conf);

#define INITIAL_DATA_SIZE 32768
#define INITIAL_VALUE_TABLE_SIZE 1024

/* Look up a variable name, and return its value string.  NULL if the
   variable doesn't exist.  If len is non-null, give the length of the
   string returned if the variable exists, don't modify otherwise.
*/
const char *
lookup_conf_in_config(const char *name, struct configuration *conf)
{
  ENTRY ent, *entp = &ent;

  ent.key = (char *) name;

  entp = hsearch(ent, FIND, conf->htab);
  if (entp != NULL)
    return( ((struct value *) entp->data)->value );
  else {
    fprintf(stderr,
            "%s: lookup_conf: lookup of config option \"%s\" failed.\n",
            progname, name);
    return NULL;
  }
}

static const char *
lookup_conf_with_fallback_in_config(const char *name, const char *fallback,
                                    struct configuration *conf)
{
  ENTRY ent, *entp = &ent;

  ent.key = (char *) name;

  entp = hsearch(ent, FIND, conf->htab);
  if (entp != NULL)
    return( ((struct value *) entp->data)->value );
  else {
    ent.key = (char *) fallback;
    entp = hsearch(ent, FIND, conf->htab);
    if (entp != NULL)
      return( ((struct value *) entp->data)->value );
    else {
      fprintf(stderr,
              "%s: lookup_conf_with_fallback: lookup of config options \"%s\" and \"%s\" both failed.\n",
              progname, name, fallback);
      return NULL;
    }
  }
}

static const char *
lookup_conf_with_numeric_in_config(const char *name,
                                   struct configuration *conf,
                                   int *numeric)
{
  ENTRY ent, *entp = &ent;

  ent.key = (char *) name;

  entp = hsearch(ent, FIND, conf->htab);
  if (entp != NULL) {
    *numeric = ((struct value *) entp->data)->numeric;
    return( ((struct value *) entp->data)->value );
  } else {
    fprintf(stderr,
            "%s: lookup_conf: lookup of config option \"%s\" failed.\n",
            progname, name);
    return NULL;
  }
}

static const char *
lookup_conf_no_error_in_config(const char *name,
                               struct configuration *conf)
{
  ENTRY ent, *entp = &ent;

  ent.key = (char *) name;

  entp = hsearch(ent, FIND, conf->htab);
  if (entp != NULL)
    return( ((struct value *) entp->data)->value );
  else
    return NULL;
}

const char *
lookup_conf(const char *name)
{
  return lookup_conf_in_config(name, config);
}

const char *
lookup_conf_with_numeric(const char *name, int *numeric)
{
  return lookup_conf_with_numeric_in_config(name, config, numeric);
}

const char *
lookup_conf_with_fallback(const char *name, const char *fallback)
{
  return lookup_conf_with_fallback_in_config(name, fallback, config);
}

const char *
lookup_conf_no_error(const char *name)
{
  return lookup_conf_no_error_in_config(name, config);
}

/* Make a new entry if necessary, otherwise return the old one.  Tell us
   which this is.  Caller's responsibility to ensure sufficient space
   in values[] and data[].
*/
static struct value *
get_or_new_value(const char *name, int *new_value, struct configuration *conf)
{
  ENTRY ent, *entp;
  int name_len = strlen(name);

  strcpy(conf->data_ptr, name);
  ent.key = conf->data_ptr;
  conf->data_ptr += name_len + 1;
  ent.data = (char *) (conf->value_ptr);

  entp = hsearch(ent, ENTER, conf->htab);
  if ( entp == NULL ) { if (new_value) *new_value = -1; return NULL; }
  if ( entp->data != ent.data ) {
    /* The variable was already in the table. */
    /* Back off the addition of the variable name to the string
       database. */
    conf->data_ptr -= name_len + 1;
    if (new_value) *new_value = 0;
    return (struct value *) (entp->data);
  } else {
    /* The variable is new. Mark the value as in-use. */
    conf->value_ptr++;
    if (new_value) *new_value = 1;
    return (struct value *) (entp->data);
  }
}

/* Set a variable, pre-existing or not. -errno on error. */
int
set_conf_in_config(const char *name, const char *val,
                   struct configuration *conf)
{
  int val_len;
  struct value *valp;
  int new;

  val_len = strlen(val);

  if ((conf->data_ptr + strlen(name) + val_len + 2) >= conf->data_end ||
       conf->value_ptr >= conf->values_end) {
    /* Not enough memory */
    /* This may reject a few cases where things would actually fit, but
       it shouldn't be a big deal.
       */
    /* XXX: expand the tables if possible */
    return(-ENOMEM);
  }

  valp = get_or_new_value(name, &new, conf);
  if (valp == NULL) return (-ENOMEM); /* Out of memory */
  if (!new && val_len <= valp->value_len) {
    /* The variable was already in the table, and there is room in the
       current space used by the database for the new value. */
    strcpy(valp->value, val);
    /* entp->val_len stays the same, so we can grow back up to the
       original size. */
  } else {
    /* It's new, or it won't fit. */
    /* XXX In the latter case, we leak database space here.  Garbage
       collection? */
    strcpy(conf->data_ptr, val);
    valp->value = conf->data_ptr;
    valp->value_len = val_len;
    valp->numeric = numeric_for(name);
    conf->data_ptr += val_len + 1;
  }
  return 0;
}

/* (Re)load the configuration file.  -errno on system error, <line> on error
   reading the system file, 0 on success.
   If we're re-loading, leave the old configuration in place until the new
   one is ready.
*/

int
load_config(void)
{
  struct configuration *new_conf;
  FILE *config_file;
  int retval;

  retval = init_config(&new_conf);
  if (retval)
    return retval;

  config_file = fopen(CONFIG_FILE, "r");
  if (config_file == NULL)
    return(-errno);

  retval = scan_config(config_file, new_conf);
  if (retval)
    return retval;

  retval = fclose(config_file);
  if (retval)
    return(-errno);

  retval = setup_client_chars(new_conf);
  if (retval)
    return retval;

  retval = setup_regexps(new_conf);
  if (retval) {
    if (retval > 0)
      fprintf(stderr,
              "%s: setup_regexps: failed to scan regexp (error %d).\n",
              progname,
              retval);
    return(-EINVAL);
  }

  if (config != NULL) {
    free_config(config);
    fprintf(stderr,
            "%s: load_config: configuration file reloaded.\n",
            progname);
  }
  config = new_conf;

  return 0;
}

int
set_conf(const char *name, const char *val)
{
  return(set_conf_in_config(name, val, config));
}

static int
setup_client_chars(struct configuration *conf)
{
  const char *s;

  s = lookup_conf_in_config("client_chars", conf);
  if(!s || strlen(s) < 4) {
    printf("chatserv: error getting client escape characters\n");
    return(-1);
  }

  strncpy(conf->client_chars, s, 4);
  return(0);
}

static int
setup_regexps(struct configuration *conf)
{
  int error, retval = 0;
  const char *pose_str, *dice_str, *ban_str, *badn_str, *alia_str, *narr_str,
      *spoof_str;

  pose_str = lookup_conf_in_config("pose_regexp", conf);
  dice_str = lookup_conf_in_config("roll_regexp", conf);
  ban_str  = setup_ban_regexp(conf);
  badn_str = lookup_conf_in_config("badnick_regexp", conf);

#ifdef DO_ALIASES
  alia_str = lookup_conf_in_config("alias_say_regexp", conf);
  narr_str = lookup_conf_in_config("narrate_regexp", conf);
  if(!alia_str || !narr_str) return -1;
#endif

#ifdef DO_SPOOF
  spoof_str = lookup_conf_in_config("spoof_regexp", conf);
  if(!spoof_str) return -1;
#endif

  if (!pose_str || !dice_str || !ban_str || !badn_str) return -1;

  error = regcomp(conf->regexps + POSE_REGEX,
                    pose_str,
                    REG_EXTENDED|REG_NOSUB);
  if (error) retval = error;

  error = regcomp(conf->regexps + DICE_REGEX,
                  dice_str,
                  REG_EXTENDED);
  if (error) retval = error;

  error = regcomp(conf->regexps + BAN_REGEX,
                  ban_str,
                  REG_EXTENDED|REG_NOSUB);
  if (error) retval = error;

  error = regcomp(conf->regexps + BADNICK_REGEX,
                  badn_str,
                  REG_EXTENDED|REG_NOSUB);
  if (error) retval = error;

#ifdef DO_ALIASES
  error = regcomp(conf->regexps + ALIAS_REGEX,
                  alia_str,
                  REG_EXTENDED|REG_NOSUB);
  if (error) retval = error;

  error = regcomp(conf->regexps + NARRATE_REGEX,
                  narr_str,
                  REG_EXTENDED|REG_NOSUB);
  if (error) retval = error;
#endif

#ifdef DO_SPOOF
  error = regcomp(conf->regexps + SPOOF_REGEX,
                  spoof_str,
                  REG_EXTENDED);
  if (error) retval = error;
#endif

  return retval;
}

/* Set up the regular expression for banned hosts, from the hosts_banned
   glob pattern. */
static char *
setup_ban_regexp(struct configuration *conf)
{
  static dyn_string ban_pattern = DYN_STR_INIT(128), *p;
  const char *banned_hosts = lookup_conf_in_config("hosts_banned", conf);
  char small_buf[2] = "A";

  p = dyn_strcpy(&ban_pattern, "");
  if (p == NULL) return NULL;

  if (banned_hosts == NULL) {
    dyn_strcpy(&ban_pattern, "^$");
    return(ban_pattern.buf);
  }

  p = dyn_strcat(&ban_pattern, "^(");
  if (p == NULL) return NULL;
  for (; *banned_hosts != '\0'; banned_hosts++) {
    if (isspace(*banned_hosts)) {
      while (isspace(*banned_hosts))
        banned_hosts++;
      if (*banned_hosts == '\0')
        break;
      else {
        p = dyn_strcat(&ban_pattern, "|");
        if (p == NULL) return NULL;
        banned_hosts--;
      }
    } else if (*banned_hosts == '.') {
      p = dyn_strcat(&ban_pattern, "\\.");
      if (p == NULL) return NULL;
    } else if (*banned_hosts == '*') {
      p = dyn_strcat(&ban_pattern, "[^.]+");
      if (p == NULL) return NULL;
    } else {
      *small_buf = *banned_hosts;
      p = dyn_strcat(&ban_pattern, small_buf);
      if (p == NULL) return NULL;
    }
  }
  p = dyn_strcat(&ban_pattern, ")$");
  if (p == NULL) return NULL;
  return ban_pattern.buf;
}

/* Initialize a configuration. -errno on error, else 0. */
static int
init_config(struct configuration **confp)
{
  struct configuration *conf;
  int retval;

  *confp = (struct configuration *) malloc(sizeof(struct configuration));
  if (*confp == NULL)
    return -ENOMEM;
  conf = *confp;

  conf->data = (char *) malloc(INITIAL_DATA_SIZE);
  if (conf->data == NULL) {
    free(conf);
    return -ENOMEM;
  }
  conf->data_ptr = conf->data;
  conf->data_end = conf->data + INITIAL_DATA_SIZE;
  conf->data_size = INITIAL_DATA_SIZE;

  conf->values = (struct value *)
    malloc(sizeof(struct value) * INITIAL_VALUE_TABLE_SIZE);
  if (conf->values == NULL) {
    free(conf->data);
    free(conf);
    return -ENOMEM;
  }
  conf->value_ptr  = conf->values;
  conf->values_end = conf->values + INITIAL_VALUE_TABLE_SIZE;
  conf->value_table_size = INITIAL_VALUE_TABLE_SIZE;

  conf->htab = calloc(1, sizeof(struct hsearch_data));
  if (conf->htab == NULL) {
    free(conf->values);
    free(conf->data);
    free(conf);
    return -ENOMEM;
  }

  retval = hcreate(INITIAL_VALUE_TABLE_SIZE, conf->htab);
  if (!retval) {
    free(conf->htab);
    free(conf->values);
    free(conf->data);
    free(conf);
    return -ENOMEM;
  }
  return 0;
}

static void
free_config(struct configuration *conf)
{
  regfree(conf->regexps + POSE_REGEX);
  regfree(conf->regexps + DICE_REGEX);
  regfree(conf->regexps + BAN_REGEX);
  regfree(conf->regexps + BADNICK_REGEX);
  hdestroy(conf->htab);
  free(conf->htab);
  free(conf->values);
  free(conf->data);
  free(conf);
}

/* Get a particular regexp. */
regex_t *
get_conf_regex(regex_index index)
{
  return config->regexps + index;
}

/* Get the client characters. */
const char *
get_conf_client_chars(void)
{
  return config->client_chars;
}

/* Reload the configuration file. */
int reload_cmd(const char *cmd_name UNUSED, struct connection *user,
               char *args UNUSED)
{
  int retval;

  retval = load_config();

  if (retval) {
    if (retval < 0)
      fprintf(stderr,
              "%s: error while reading config file: %s.\n",
              progname, strerror(-retval));
    else if (retval > 0)
      fprintf(stderr,
              "%s: syntactic error in config file, line %d.\n",
              progname, retval);
    return(write_to_user(formatn("reload_failed", user, NULL),
                         user));
  }

  reset_channel_topics();
  report_client_formats();

  return(write_to_user(formatn("reload_succeeded", user, NULL),
                       user));
}
