/*
 * Calvin chatserver connection management.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 1995 Jonathan Lennox
 * Copyright 2017 Russ Allbery <eagle@eyrie.org>
 *
 * See LICENSE for licensing terms.
 */

/* AIX requires this to be the first thing in the file... */
#if !defined (__GNUC__) && defined (_AIX)
 #pragma alloca
#endif

#include <config.h>
#include <portable/socket.h>
#include <portable/system.h>

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <regex.h>
#include <signal.h>
#include <time.h>

#ifdef __GNUC__
# undef alloca
# define alloca(size) __builtin_alloca(size)
#else
# if HAVE_ALLOCA_H
#  include <alloca.h>
# else
#  if !defined(_AIX) && !defined(alloca) /* predefined by HP cc +Olibcalls */
char *alloca ();
#  endif
# endif
#endif

#ifndef O_NONBLOCK
#define O_NONBLOCK O_NDELAY
#endif

#include <util/macros.h>

#include "channels.h"
#include "connections.h"
#include "format.h"
#include "config-file.h"
#include "commands.h"
#include "nicks.h"
#include "telnet.h"
#include "lastlog.h"

extern char *progname;
extern int dns_fd;

/* Global variables about connections */
struct connection *connections = NULL;
int reap = 0;

int dns_child; /* XXX pid_t? */

/* This is because write_wrapped requires that its argument be
   writable, and constant strings aren't by default (I don't want to
   have to compile with -fwritable-strings).
*/
char blankline[] = "\n";

/* Declarations  for static functions */
static int dispatch_user_line(char *line, struct connection *user);
static void dns_lookup_loop(int fd);
static void do_dns_lookup(struct connection *newcon);

/* Functions about connections. */

/* Watch for incoming TCP connections on the given port.  Return
   fd; -errno on error. */
int tcp_listen_on(int port)
{
  int fd, error, opt, flags;
  struct sockaddr_in service;

  /* The basic TCP socket */
  fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd < 0) return -errno;

  /* Standard TCP options */
  opt = 1;
  error = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *)&opt,
                     sizeof(opt));
  if (error < 0) return -errno;
  error = setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (char *)&opt,
                     sizeof(opt));
  if (error < 0) return -errno;
  /* XXX Do I want SO_LINGER also? */

  flags = fcntl(fd, F_GETFL, 0);
  if (flags == -1) return -errno;
  error = fcntl(fd, F_SETFL, flags | O_NONBLOCK);
  if (error == -1) return -errno;

  service.sin_family = AF_INET;
  service.sin_addr.s_addr = INADDR_ANY;
  service.sin_port = htons(port);

  error = bind(fd, (struct sockaddr *) &service, sizeof(service));
  if (error < 0) return -errno;

  error = listen(fd, 5);
  if (error < 0) return -errno;

  return fd;
}

/* If the given host is a banned host. */
static int banned_host(char *hostname)
{
  if (regexec(get_conf_regex(BAN_REGEX), hostname, 0, NULL, 0) == 0)
    return 1;
  else return 0;
}

/* We've selected the incoming connection port for reading!  Set up a
   new user connection.  Return the connection, NULL on error,
   preserving errno.*/
struct connection * setup_new_connection(int fd)
{
  struct connection * newcon;
  socklen_t len = sizeof(struct sockaddr_in);
  int flags, error;

  newcon = (struct connection *)malloc(sizeof(struct connection));
  if (newcon == NULL) return NULL;

  newcon->fd = accept(fd, (struct sockaddr *) &(newcon->connection), &len);
  if (newcon->fd < 0) {
    free(newcon);
    return NULL;
  }
  /* Set up the values of the connection struct. */
  error = init_connection_channels(newcon);
  if (error) return NULL;

  /* BSD 4.4 and Solaris inherit non-blocking status from the socket "fd"
     refers to.  Linux and SunOS don't.  So we set it again, here.
     Aaaaaaaaaaaaaaaargh. */
  flags = fcntl(newcon->fd, F_GETFL, 0);
  if (flags == -1) return NULL;
  error = fcntl(newcon->fd, F_SETFL, flags | O_NONBLOCK);
  if (error == -1) return NULL;

  newcon->hostname_resolved = 0;

  /* Only various fprintf error messages should use this, I think */
  newcon->hostname = strdup(inet_ntoa(newcon->connection.sin_addr));

  if (banned_host(inet_ntoa(newcon->connection.sin_addr))) {
    close(newcon->fd);
    free(newcon->hostname);
    free(newcon);
    errno = ECONNREFUSED;
    return NULL;
  }

#if DO_UNORTHODOX
  newcon->unorthoname = NULL;
  newcon->unorthodox = newcon->unorthocount = 0;
#endif
  newcon->nick = (char *) malloc(11);
  if (newcon->nick == NULL) {
    free(newcon->hostname);
    free(newcon);
    return NULL;
  }
  strcpy(newcon->nick, "Logging_on");

  newcon->reap_me = 0;
  newcon->columns = 0;
  newcon->print_blank = 0;
  newcon->client_flags = 0;
  newcon->buffer[0] = '\0';
  newcon->buf_end = 0;
  newcon->has_overflowed = 0;
  newcon->privs = logging_on_priv;
  newcon->status = unnamed;
  newcon->last_nonidle = newcon->signon = time(NULL);
  newcon->target.argtype = 0;
  newcon->error = NULL;
  newcon->waiting = 0;

  /* Add the new connection to the linked list. */
  newcon->prev = NULL;
  newcon->next = connections;
  if (connections != NULL) connections->prev = newcon;
  connections = newcon;

  /* Put in a request for the new connection's DNS resolution. */
  do_dns_lookup(newcon);

  newcon->ssl = (SSL) 0;
  return newcon;
}

#ifdef DO_SSL

/* Set up a new SSL connection. */
struct connection * setup_new_ssl_connection(int fd)
{
  struct connection * newcon = setup_new_connection(fd);
  return start_ssl_setup(newcon, newcon->fd);
}

#endif

/* Close a connection, after a drop or a signoff or whatever. */
int close_connection(struct connection *user)
{
  /* We do the actual connection-closing in a post-select-loop loop, to
     prevent linked-list-traversals from getting upset. */
  /* Make sure no one can see them or write to them. */
  /* We distinguish between how fully visible people die and how the
     unconnected die. */
  if (user->status == normal || user->status == gone)
    user->status = gone;
  else
    user->status = gone_too_soon;
  /* Stop them from executing commands. */
  user->privs = no_priv;
  /* Make sure they can't write to a channel. */
  user->channels.primary_channel = -1;
  /* Kill this connection in the post-select-loop loop. */
  user->reap_me = 1;
  /* Run the post-select-loop loop, later. */
  reap = 1;
  return 0;
}

/* Try to clear out this user's buffer. */
void attempt_write(struct connection *user)
{
  int r;
#ifdef DO_SSL
  if(user->ssl)
    r = ssl_write(user->ssl, user->output_buffer, user->waiting);
  else
#endif
    r = write(user->fd, user->output_buffer, user->waiting);

  if (r <= 0) {
    char *error = strdup(r ? strerror(errno) : "No data written");

    fprintf(stderr, "%s: attempt_write: connection %d (%s from %s): "
            "error writing: %s.\n",
            progname, user->fd, nickname(user), user->hostname, error);
    close_connection(user);
    user->error = error;
    return;
  }

  user->waiting -= r;
  if(user->waiting) {
    memmove(user->output_buffer, user->output_buffer + r, user->waiting);
  }
}

/* Queue this up to be written to the user. */
int queue_for_write(struct connection *user, const char *string, int len)
{
  if(user->waiting + len > OUTPUT_BUFSIZE) {
    close_connection(user);
    user->error = strdup("Flooded");
    return -1;
  }

  memcpy(user->output_buffer + user->waiting, string, len);
  user->waiting += len;
  return len;
}

/* Write to a connection, wrapping at the appropriate column and
   sending \r\n for \n
   Note that we go through some convolutions to ensure that the \r\n
   is sent in the same write() command as the associated text.  Bad
   things can happen over laggy connections otherwise.
*/
static int write_wrapped(struct connection *user, size_t columns, char *string,
                         size_t string_len)
{
  int retval, chars_written = 0;
  size_t i, line_start, last_good_break;
  int wrote_newline;
  char saved[2];

  if (columns == 0) columns = INT_MAX;

  line_start = last_good_break = 0;

  for (i = 0; i < string_len + 1; i++) {
    for (; i < string_len + 1 && i - line_start < columns; i++) {
      if (string[i] == '\n') {
        last_good_break = i;
        break;
      }
#if 0
      else if (string[i] == '\0') {
        last_good_break = i - 1;
        break;
      }
#endif
      else if (string[i] == ' ') {
        last_good_break = i;
        if (i == line_start && i != 0) {
          line_start++;
          last_good_break++;
        }
      }
    }
    if (last_good_break == line_start)
      last_good_break = i;

    if (i >= string_len)
      last_good_break = string_len;

    wrote_newline = 0;

    if (last_good_break > line_start) {
      if (i < string_len) {
        wrote_newline = 1;
        saved[0] = string[last_good_break];
        saved[1] = string[last_good_break+1];
        string[last_good_break]   = '\r';
        string[last_good_break+1] = '\n';
      }
      retval = queue_for_write(user, &(string[line_start]),
                               last_good_break - line_start
                               + (wrote_newline?2:0));
      if (retval < 0) {
        if (wrote_newline) {
          string[last_good_break] = saved[0];
          string[last_good_break+1] = saved[1];
        }
        return retval;
      }
      chars_written += retval;
      if (wrote_newline) {
        string[last_good_break] = saved[0];
        string[last_good_break+1] = saved[1];
      }
    }
    if (!wrote_newline && i < string_len) {
      retval = queue_for_write(user, "\r\n", 2);
      if (retval < 0) return retval;
    }

    line_start = last_good_break;

    if (string[line_start] == '\n') {
      line_start++;
      chars_written++;
    }
    else while (line_start < string_len + 1 && string[line_start] == ' ') {
      line_start++;
      chars_written++;
      if (i < line_start) i = line_start;
    }

    last_good_break = line_start;
  }

  return chars_written;
}

/* Announce the connection of a user. */
int announce_connection(struct connection *user)
{
  int retval;

  retval = insert_to_log(formatn("signon_ll_entry", user, NULL));
  if (retval) return retval;
  return write_to_everyone_checkspam(formatn("signon", user, NULL), 0);
}

/* Announce the (impending?) disconnection of a user. */
int announce_disconnection(struct connection *user, const char *reason)
{
  int retval;
  if (user->status == normal || user->status == gone) {
    /* Don't announce the disconnection of someone who hasn't completed
       the login process, except to that user. */
    retval = insert_to_log(formatn("signoff_ll_entry", user,
                                  NULL));
    if (retval) return retval;
    return write_to_everyone_checkspam(formatn("signoff", user, reason),
                                       user->channels.channel_map);
  }
  else if (user->status != unnamed) {
    /* Since the signoff message includes nick, don't bother if they haven't
       got one yet. */
    return write_to_user(formatn("signoff", user, reason),
                         user);
  }
  return 0;
}

/* The user took an action that is visible to other users on the server.
   These are the consequences of this.
   -errno on errors.
*/
int visible_action(struct connection *user)
{
  int retval;

  if (user->status == unrevealed) {
    user->status = normal;
    retval = announce_connection(user);
    if (retval) return retval;
    retval = write_to_user(formatvn("now_visible", user, 0), user);
    if (retval) return retval;
  } else if (user->status != normal && user->status != cloaked) {
    user->status = normal;
    retval = write_to_user(formatvn("now_visible", user, 0), user);
    if (retval) return retval;
  }

  user->last_nonidle = time(NULL);

#if DO_UNORTHODOX
  if (user->unorthodox) {
    user->unorthodox = (user->unorthodox + 1) % (user->unorthocount + 1);
    if(user->unorthodox == 0) user->unorthodox++;
  }
#endif

  return 0;
}

/* There's input from a user's connection -- read it, and dispatch it as
   appropriate. EOF on connection close, errno on errors.*/
int read_from_user(struct connection *user)
{
  int retval, i, line_start, old_errno;

  retval = telnet_read(user, &(user->buffer[user->buf_end]),
                       BUFFERSIZE - user->buf_end - 1);
  if (retval == -1) {
    /* End of File */
    const char *drop_msg = lookup_conf("default_drop_msg");

    if (drop_msg == NULL) drop_msg = "";
    announce_disconnection(user, drop_msg);
    close_connection(user);
    return EOF;
  }
  else if (retval < -1) {
    /* Error */
    old_errno = errno;
    /* On a clean drop (above), announce then close.  On an error, close
       then drop, since we don't know what might be wrong with the
       connection. */
    close_connection(user);
    announce_disconnection(user, strerror(old_errno));
    return old_errno;
  }
  user->buffer[user->buf_end + retval] = '\0';

  for (line_start = 0, i = user->buf_end;
       i < BUFFERSIZE && user->buffer[i] != '\0';
       i++) {
    if (user->buffer[i] == '\n' || user->buffer[i] == '\r') {
      user->buffer[i] = '\0';
      /* Lines may end with CRLF or LFCR.  Interpret both of these as
         EOL. */
      if (i < BUFFERSIZE - 1 &&
          (user->buffer[i+1] == '\n' || user->buffer[i+1] == '\r') &&
          user->buffer[i] != user->buffer[i+1])
        user->buffer[++i] = '\0';
      if(user->has_overflowed) {
        retval = user->has_overflowed = 0;
      } else {
        retval = dispatch_user_line(&(user->buffer[line_start]), user);
      }
      if (retval) {
        if (retval == -ENOTCONN) {
          return EOF;
        } else {
          return -retval;
        }
      }
      line_start = i + 1;
    }
  }

  if (line_start != 0) {
    strcpy(user->buffer, &(user->buffer[line_start]));
    user->buf_end = i - line_start;
  }
  else {
    user->buf_end = i;
    if (i >= BUFFERSIZE - 1) {
      /* Our buffer is full of only one line.  Time to, as the metaphor
         goes, drop back 10 and punt, dispatching the incomplete line. */
      retval = dispatch_user_line(&(user->buffer[line_start]), user);
      if (retval) {
        if (retval == -ENOTCONN) {
          return EOF;
        } else {
          return -retval;
        }
      }
      user->buf_end = 0;
      user->has_overflowed = 1;
    }
  }
  return 0;
}

/* Determine if the given line is indeed a pose.  Assumes that the first
   char is a : or ;. */
static int is_pose(char *line)
{
  int retval;

  retval = regexec(get_conf_regex(POSE_REGEX), line, 0, NULL, 0);
  if (retval == 0) return 1;
  else return 0;
}

static int is_alias(char *line)
{
  int retval;

  retval = regexec(get_conf_regex(ALIAS_REGEX), line, 0, NULL, 0);
  if(retval == 0) return 1;
  else return 0;
}

static int is_narrate(char *line)
{
  int retval;

  retval = regexec(get_conf_regex(NARRATE_REGEX), line, 0, NULL, 0);
  if(retval == 0) return 1;
  else return 0;
}

#ifdef DO_SPOOF
static int check_spoof(struct connection *user, char *line)
{
  int retval, i;
  regmatch_t pmatch[9];
  char *elements[8];
  char *s;

  retval = regexec(get_conf_regex(SPOOF_REGEX), line, 9, pmatch, 0);
  if(retval == 0) {
    for(i = 1; i <= 8; i++) {
      if(pmatch[i].rm_so > -1) {
        elements[i-1] = malloc(pmatch[i].rm_eo - pmatch[i].rm_so + 1);
        if(elements[i-1] == NULL) return -ENOMEM;
        strncpy(elements[i-1], line + pmatch[i].rm_so,
                pmatch[i].rm_eo - pmatch[i].rm_so);
        elements[i-1][pmatch[i].rm_eo - pmatch[i].rm_so] = '\0';
      } else elements[i-1] = NULL;
    }
    s = formatvn("spoof_rewrite_as", user, 8,
                 elements[0], elements[1], elements[2], elements[3],
                 elements[4], elements[5], elements[6], elements[7]);
    i = strlen(line);
    strncpy(line, s, i);
    for(i = 0; i <= 7; i++) if (elements[i]) free(elements[i]);
  }
  return(0);
}
#endif

/* This isn't anywhere near a complete translation table, because there are
   two separate problematic character sets: ANSI (Windows) and Mac Roman.
   What we do is simply handle the most common cases--dashes and "smart"
   quotes--from *both* character sets.  Everything else with a high bit set,
   and all other control characters, are harmlessly turned into spaces. */

const char *translation_table =
"\
                                \
 !\"#$%&'()*+,-./0123456789:;<=>?\
@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_\
`abcdefghijklmnopqrstuvwxyz{|}~ \
                 \'\'\"\" --        \
                                \
                --\"\"\'\'          \
                                \
";

/* Handle backspaces, etc., in a line. */
static char *backspace_cleanup_and_translate(char *line)
{
  unsigned char *p;
  char *q;

  p = (unsigned char *) line;
  q = line;

  while (*p != '\0') {
    if (*p == 'H' - 'A' + 1) {
      p++;
      if (q != line) q--;
    }
    else if (*p == 'U' - 'A' + 1) {
      p++;
      q = line;
    }
    else
      *q++ = translation_table[*p++];
  }
  *q = '\0';
  return line;
}


int greet_new_user(struct connection *user)
{
  user->privs = normal_priv;
  user->status = unrevealed;
  /* Show who's logged onto the server right now, short form. */
  return users_cmd("USERS", user, (char *) "*");
}

/* Dispatch the input line. -errno on error.*/
static int dispatch_user_line(char *line, struct connection *user)
{
  int retval;

  line = backspace_cleanup_and_translate(line);

  if (user->status == unnamed) {
    if (line[0] == '/' || line[0] == '@') {
      if (line[1] != line[0]) {
        user->target.argtype = RCPT_UNSPEC;
        user->target.explicit = 0;
        retval = exec_command(line + 1, user);
        if (retval) return retval;
      } else {
        /* We begin with two /'s or @'s, so move the pointer past the
           first one and print the second.
           */
        line++;
      }
    } else if (line[0] != '\0') {
      retval = nickname_cmd("NICK", user, line);
      if (retval) return retval;
    }

    if (user->status == unnamed) {
      /* If we're still unnamed ... */
      return write_to_user(formatn("nick_request",
                                  user, NULL),
                           user);
    }
    return 0;
  }

  /* We know we're "real", now. */
  if (valid_channel(user->channels.primary_channel)) {
    user->target.argtype = RCPT_CHANNEL;
    user->target.u.channel = user->channels.primary_channel;
  } else {
    user->target.argtype = RCPT_UNSPEC;
  }
  user->target.explicit = 0;

#ifdef DO_ALIASES
  user->alias[0] = '\0';
#endif

#ifdef DO_UNORTHODOX
  if(user->unorthodox && (user->unorthoname[user->unorthodox - 1]))
    strcpy(user->alias, user->unorthoname[user->unorthodox - 1]);
#endif

  return parse_user_line(line, user);
}

int parse_user_line(char *line, struct connection *user)
{
  if (line[0] == '\0') {
    /* Empty line */
    return 0;
  }
  if (line[0] == '/' || line[0] == '@')
    if (line[1] != line[0]) {
      return exec_command(line + 1, user);
    } else {
      /* We begin with two /'s or @'s, so move the pointer past the
         first one and print the second.
         */
      line++;
    }

  /* : means pose, ; means posessive-pose. */
  else if (line[0] == ':' || line[0] == ';') {
    if (line[1] != line[0]) {
      if (is_pose(line)) {
        char c = *line;
        line++;
        while (isspace(*line)) line++;
        if (c == ':')
          return send_to_recipient(user, line, "pose", user->target);
        else /* c == ';' */
          return send_to_recipient(user, line, "ppose", user->target);
      }
    } else {
      /* We begin with two :'s or two ;'s, so move the pointer past the
         first one and print the second.
         */
      line++;
    }
  } else if (line[0] == '%') {
    if (line[1] != line[0])
      return whisper_cmd("%", user, line + 1);
    else
      line++;
  }
#ifdef DO_ALIASES
  else if (line[0] == '[') {
    if (line[1] != line[0]) {
      if (is_alias(line)) {
        if((user->target.argtype == RCPT_CHANNEL) &&
           (user->channels.channel_map & 1 << user->target.u.channel)) {
          char *l1 = line, *l2;
          l1++;
          l2 = strchr(l1, ']');
          if(l2) {
            if(l2 - l1 > MAX_NICK_LENGTH) l2 = l2 - (l2 - l1 - MAX_NICK_LENGTH);
            strncpy(user->alias, l1, l2-l1);
            user->alias[l2-l1] = '\0';
            l2 += 2;
            while(isspace(*l2)) l2++;
          } else l2 = line;
          return send_to_recipient(user, l2, "send", user->target);
        } else return send_to_recipient(user, line, "send", user->target);
      } else if (is_narrate(line)) {
        char *l1 = line, *l2;
        l1++;
        l2 = strchr(l1, ']');
        if(l2) {
          strcpy(l2, l2+1);
        }
        return send_to_recipient(user, l1, "narrate", user->target);
      }
    }
  }
#endif

  return send_to_recipient(user, line, "send", user->target);
}

int send_to_recipient(struct connection *user, char *line,
                      const char *msg_type, struct recipient rcpt)
{
  char lookup1[64], lookup2[64];
  char *formatted1, *formatted2;
  int retval;
  int old_primary;

  if (line[0] == '\0')
    return 0;
  if (strlen(line) > MAXMSGLENGTH) line[MAXMSGLENGTH] = 0;

  strcpy(lookup1, msg_type);
  strcpy(lookup2, msg_type);

#ifdef DO_SPOOF
  retval = check_spoof(user, line);
  if(retval) return retval;
#endif

  switch (rcpt.argtype) {
  case RCPT_USER:
    retval = visible_action(user);
    if (retval) return retval;

    strcat(lookup1, "_whisper_sender");
    strcat(lookup2, "_whisper_recipient");
    retval = write_to_user(format2n(lookup1,
                                   rcpt.u.user, line, nickname(user)),
                           user);

    if (retval) return retval;

    if (rcpt.u.user->print_blank) {
      retval = write_to_user(blankline, rcpt.u.user);
      if (retval) return retval;
    }
    return write_to_user(formatn(lookup2, user, line),
                         rcpt.u.user);
    break;
  case RCPT_CHANNEL:
    retval = visible_action(user);
    if (retval) return retval;
    old_primary = user->channels.primary_channel;
    user->channels.primary_channel = rcpt.u.channel;
    if (user->channels.channel_map & (1 << rcpt.u.channel)) {
      /* User is on the channel being sent to */
      strcat(lookup1, "_on_channel_same_channel");
      strcat(lookup2, "_on_channel_other_channel");
#ifdef DO_ALIASES
      if((*(user->alias)) && (
         (!strncmp(lookup1, "send", 4)) ||
         (!strncmp(lookup1, "pose", 4)) ||
         (!strncmp(lookup1, "ppose", 5)) ) ) {
        strcat(lookup1, "_alias");
        strcat(lookup2, "_alias");
      }
#endif
    } else {
      strcat(lookup1, "_off_channel_same_channel");
      strcat(lookup2, "_off_channel_other_channel");
      retval = write_to_user(formatn(lookup2,
                                    user, line),
                             user);
      if (retval) {
        user->channels.primary_channel = old_primary;
        return retval;
      }
    }
#ifdef DO_ALIASES
    formatted1 = format2n(lookup1, user, line, user->alias);
    formatted2 = format2n(lookup2, user, line, user->alias);
#else
    formatted1 = formatn(lookup1, user, line);
    formatted2 = formatn(lookup2, user, line);
#endif
    user->channels.primary_channel = old_primary;
    retval = channel_nonidle(rcpt.u.channel);
    if (retval) return retval;
    return write_to_channel(formatted1, formatted2, rcpt.u.channel);
    break;
  case RCPT_EVERYONE:
    strcat(lookup1, "_yell");
    retval = visible_action(user);
    if (retval) return retval;
    return write_to_everyone(formatn(lookup1, user, line));
    break;
  default:
    /* XXX I *think* this should only be RCPT_NONE...just drop it. */
    break;
  }
  return 0;
}

/* Write a given string to the given user's connection. -errno on
   error. */
/* LF -> CRLF translations should have been done already. */
int write_to_user(char *string, struct connection *user)
{
  int len;

  if(!(user->client_flags & CLIENT_NUMERICS)) {
    string = strdup(string);
    if(!string) return(-1);
    (void) strip_for_user(string);
  }

  if (user->reap_me)
    return -ENOTCONN; /* XXX Is this correct? */

  len = strlen(string);
  write_wrapped(user, user->columns, string, len);
  /* Do something with the result?  Or just get rid of it? */

  if(!(user->client_flags & CLIENT_NUMERICS)) free(string);
  return 0;
}

/* Write a given string to everyone.  -errno on error. */
int write_to_everyone(char *string)
{
  struct connection *p;
  int retval, old_retval = 0;

  for (p = connections; p != NULL; p = p->next) {
    if (p->status == normal || p->status == hiding || p->privs > normal_priv) {
      /* Don't write to people in the process of logging on, unless they're
         privileged. */
      if (p->print_blank) {
        retval = write_to_user(blankline, p);
        if (retval) {
          old_retval = retval;
          continue;
        }
      }
      retval = write_to_user(string, p);
      if (retval) old_retval = retval;
    }
  }

  return old_retval;
}

#ifdef DO_SPAM
int write_to_everyone_checkspam(char *string, int map)
{
  struct connection *p;
  int retval, old_retval = 0;

  for (p = connections; p != NULL; p = p->next) {
    if ((p->status == normal || p->status == hiding || p->privs > normal_priv)
        && ((!(p->client_flags & CLIENT_NOSPAM)) ||
            (map & p->channels.channel_map) ) ) {
      /* Don't write to people in the process of logging on, unless they're
         privileged. */
      if (p->print_blank) {
        retval = write_to_user(blankline, p);
        if (retval) {
          old_retval = retval;
          continue;
        }
      }
      retval = write_to_user(string, p);
      if (retval) old_retval = retval;
    }
  }

  return old_retval;
}
#endif

/* The signal handler */
void handler(int sig)
{
  int retval;
  struct sigaction action;

  switch (sig) {
  case SIGTERM:
    /* First we need to reset our own SIGTERM behavior. */
    /* (I'd be nice to just use SA_RESETHAND, but apparently it's not Posix. */
    action.sa_handler = SIG_DFL;
    sigemptyset(&action.sa_mask);
    action.sa_flags = 0;
    retval = sigaction(SIGTERM, &action, NULL);
    if (retval)
      fprintf(stderr, "%s: handler: sigaction failed: "
              "%s.\n", progname, strerror(errno));

    /* Kill the child and then ourselves. */
    retval = kill(dns_child, SIGTERM);
    if (retval == -1)
      fprintf(stderr, "%s: handler: kill(dns_child==%d, SIGTERM) failed: "
              "%s.\n", progname, dns_child, strerror(errno));
    retval = kill(getpid(), SIGTERM);
    if (retval == -1)
      fprintf(stderr, "%s: handler: kill(getpid(), SIGTERM) failed: "
              "%s.\n", progname, strerror(errno));
    break;
  default:
    fprintf(stderr,
            "%s: handler: Received signal %d, not expecting; "
            "very confused, ignoring.\n",
            progname, sig);
    break;
  }
  return;
}


/* Commands */

/* Sign off on purpose. */
int bye_cmd(const char *cmd_name UNUSED, struct connection *user, char *args)
{
  int retval;
  const char *message = args;

  if (strcmp(message, "") == 0)
    message = lookup_conf("default_signoff_msg");

  if (!message) message = "";

  retval = announce_disconnection(user, message);
  if (retval) return retval;

  retval = close_connection(user);
  if (retval) return retval;
  return -ENOTCONN;  /* To stop anything from trying to write to this
                        connection subsequently. */
}

int pose_cmd(const char *cmd_name UNUSED, struct connection *user, char *args)
{
  return send_to_recipient(user, args, "pose", user->target);
}

int ppose_cmd(const char *cmd_name UNUSED, struct connection *user, char *args)
{
  return send_to_recipient(user, args, "ppose", user->target);
}

#ifdef DO_YAY
int yay_cmd(const char *cmd_name UNUSED, struct connection *user, char *args)
{
  char *ptr, *yay_str, *yay_str_dup;

  for (ptr = args; *ptr != '\0' && !isspace(*ptr); ptr++)
    ; /* Do nothing */

  if (*ptr != '\0') {
    *ptr = '\0';
    ptr++;
  }

  for (; *ptr != '\0' && isspace(*ptr); ptr++)
    ; /* Do nothing */

  if (*ptr != '\0') {
    yay_str = format2n("yay_format", user, args, ptr);
    yay_str_dup = alloca(strlen(yay_str) + 1);
    if (yay_str_dup == NULL) return -ENOMEM;
    strcpy(yay_str_dup, yay_str);
    return send_to_recipient(user, yay_str_dup, "send", user->target);
  }
  return 0;
}
#endif

#ifdef DO_YELL
int yell_cmd(const char *cmd_name UNUSED, struct connection *user, char *args)
{
  user->target.argtype = RCPT_EVERYONE;
  user->target.explicit = 1;
  return parse_user_line(args, user);
}
#endif

#ifdef DO_ALIASES

int as_cmd(const char *cmd_name UNUSED, struct connection *user, char *args)
{
  char *rest, *alias, *aliasend, quotechar;

  alias = args;
  quotechar = *alias;
  if(isalnum(quotechar)) {
    aliasend = strchr(alias, ' ');
  } else {
    alias++;
    aliasend = strchr(alias, quotechar);
  }
  if(aliasend) {
    *aliasend = '\0';
    strncpy(user->alias, alias, MAX_NICK_LENGTH);
    rest = aliasend + 1;
    while(*rest == ' ') rest++;
  } else {
    return write_to_user(formatn("bad_as", user, args),
                         user);
  }

  return parse_user_line(rest, user);
}

int narrate_cmd(const char *cmd_name UNUSED, struct connection *user,
                char *args)
{
  return send_to_recipient(user, args, "narrate", user->target);
}

#endif

/* Whisper to a user, or send to another channel */
int whisper_cmd(const char *cmd_name UNUSED, struct connection *user,
                char *args)
{
  char *rest;
  struct recipient rcpt;

#ifdef DO_YELL
  rcpt = identify_recipient(RCPT_USER | RCPT_CHANNEL | RCPT_EVERYONE,
                            user, args, &rest);
#else
  rcpt = identify_recipient(RCPT_USER | RCPT_CHANNEL,
                            user, args, &rest);
#endif

  if (rcpt.argtype == 0)
    return write_to_user(formatn("unknown_user", user, args),
                         user);

  user->target = rcpt;

  return parse_user_line(rest, user);
}

int wrap_cmd(const char *cmd_name UNUSED, struct connection *user, char *args)
{
  int width;

  if (strcmp(args, "") != 0) {
    width = atoi(args);
    if(user->client_flags & CLIENT_NUMERICS) width = 0;
    if (abs(width) >= 1024) {
      return write_to_user(formatn("bad_wrap", user, args),
                           user);
    }
    if (user->columns != abs(width) || user->print_blank != (width > 0)) {
      user->columns = abs(width);
      user->print_blank = (width > 0);
      if (user->columns != 0)
        return write_to_user(formatn("new_wrap", user, NULL),
                             user);
      else
        return write_to_user(formatn("new_no_wrap", user, NULL),
                             user);
    }
  }
  /* Both the case where the wrap didn't change, and the case where there
     was no argument to the function. */
  if (user->columns != 0)
    return write_to_user(formatn("current_wrap", user, NULL),
                         user);
  else
    return write_to_user(formatn("no_wrap", user, NULL),
                         user);
}

int report_client_numeric_info(struct connection *user)
{
  /* This should hold anything remotely sane */
  static char optbuf[1024]; int i;

  if(user->client_flags & CLIENT_NUMERICS) {
    sprintf(optbuf, "3901 client format characters %s end\n",
            get_conf_client_chars());
    i = write_to_user(optbuf, user);
    if(i) return(i);
    sprintf(optbuf, "3902 client format date %s end\n",
            lookup_conf("date_format"));
    i = write_to_user(optbuf, user);
    if(i) return(i);
    sprintf(optbuf, "3903 client format roll %s end\n",
            lookup_conf("roll_format"));
    i = write_to_user(optbuf, user);
    if(i) return(i);
  }
  return(0);
}

int report_client_formats(void)
{
  int error, old_error = 0;
  struct connection *user = connections;

  while(user) {
    if(user->client_flags & CLIENT_NUMERICS) {
      error = report_client_numeric_info(user);
      if(error) old_error = error;
    }
    user = user->next;
  }
  return old_error;
}

int client_cmd(const char *cmd_name UNUSED, struct connection *user,
               char *args)
{
  int i;
  /* Should be big enough to list all options */
  static char optbuf[1024];

  if(isdigit(*args))
  {
    i = (int) strtol(args, (char **)NULL, 0);
    user->client_flags = i;
    i = report_client_numeric_info(user);
    if(i) return(i);
    if((user->client_flags & CLIENT_NUMERICS) && user->columns) {
      i = wrap_cmd("WRAP", user, (char *) "0");
      if(i) return(i);
    }
  } else if(*args)
    return write_to_user(formatn("bad_client", user, NULL), user);

  *optbuf = '\0';

#ifdef DO_SPAM
  if(user->client_flags & CLIENT_NOSPAM) strcat(optbuf, " (NoSpam)");
#endif
  if(user->client_flags & CLIENT_NODEFAULT) strcat(optbuf, " (NoDefault)");
  if(user->client_flags & CLIENT_NUMERICS) strcat(optbuf, " (Numerics)");

  if(*args)
    return write_to_user(formatn("client_now", user, optbuf), user);
  else
    return write_to_user(formatn("client_current", user, optbuf), user);
}

/* Go away.  Should only be usable by priv level 2 or higher. */
int kick_cmd(const char *cmd_name UNUSED, struct connection *user, char *args)
{
  int retval;
  struct recipient target;
  char *msg, *msg_dup;

  target = identify_recipient(RCPT_USER | RCPT_UNSPEC,
                              user, args, NULL);

  if (target.argtype != RCPT_USER) {
    return write_to_user(formatn("unknown_user", user, args),
                         user);
  }

  msg = formatn("kick_msg", user, NULL);
  msg_dup = alloca(strlen(msg) + 1);
  if (msg_dup == NULL) return -ENOMEM;
  strcpy(msg_dup, msg);

  retval = announce_disconnection(target.u.user, msg_dup);
  if (retval) return retval;

  return close_connection(target.u.user);
}


/*******************************/

/* The Async DNS module stuff. */

/* The stuff that gets passed back and forth to the DNS-lookup child. */
struct dns_child_info {
  int connection_fd;                  /* Set by parent */
  time_t connection_time;             /* Set by parent */
  struct in_addr connection_addr;     /* Set by parent */
  int connection_resolved;            /* Set by child */
  char connection_hostname[64];       /* Set by child */
};

/* Start the DNS child. */

/* Cygnus doesn't have socketpair(), so we can't do reverse-DNS right. */

#ifdef __CYGWIN32__

int start_dns() { return 0; }
static void dns_lookup_loop(int fd) {}
void do_dns_lookup(struct connection *newcon) {}
void do_dns_received(int fd) {}

#else

int start_dns(void)
{
  int dnsfds[2]; /* fd pair (produced by socketpair) for reading/writing
                    0 is the one the parent uses; 1 is the child's. */
  int retval;

  retval = socketpair(AF_UNIX, SOCK_STREAM, 0, dnsfds);
  if (retval == -1)
    return -errno;

  retval = fork();
  if (retval < 0)
    return -errno;
  else if (retval == 0) {
    /* We are the child */
    /* We now try not to exit for any reason */
    retval = close(dnsfds[0]);
    if (retval)
      fprintf(stderr, "DNS Child: dns_init: close(%d) failed: %s\n",
              dnsfds[0], strerror(errno));
    dns_lookup_loop(dnsfds[1]);
    /* NOTREACHED */
    fprintf(stderr, "DNS Child: dns_init: dns_lookup_loop finished???  exiting...\n");
    exit(1);
  }
  else {
    /* retval > 0, We are the parent */
    dns_child = retval;
    close(dnsfds[1]);
    return dnsfds[0];
  }
}

/* This must be greater than sizeof(struct dns_child_info) */
#define DNS_LOOKUP_BUFFERSIZE 128

static void dns_lookup_loop(int fd)
{
  int retval;
  char *buffer = NULL;
  struct dns_child_info *info;
  struct hostent *host;

  buffer = (char *)malloc(sizeof(struct dns_child_info));
  if (buffer == NULL) {
    fprintf(stderr, "DNS Child:  Aieee!  Malloc failure.\n");
    exit(1);
  }
  info = (struct dns_child_info *) buffer;
  /* We can now use info and buffer interchangably. */

  while(1) {
    retval = read(fd, buffer, sizeof(struct dns_child_info));
    if (retval < 0) {
      fprintf(stderr, "DNS Child:  Error reading from parent: %s.\n",
              strerror(errno));
      continue;
    }

    host = gethostbyaddr((char *)&(info->connection_addr),
                         sizeof(info->connection_addr), AF_INET);
    if (host == NULL) {
      info->connection_resolved = 0;
      info->connection_hostname[0] = '\0';
    }
    else {
      info->connection_resolved = 1;
      strncpy(info->connection_hostname, host->h_name, 63);
      info->connection_hostname[63] = '\0';
    }

    retval = write(fd, buffer, sizeof(struct dns_child_info));
    if (retval < 0) {
      fprintf(stderr, "DNS Child: Error writing to parent: %s.\n",
              strerror(errno));
    }
    /* Loop forever. */
  }
  /* Not reached. */
}


/* Send the DNS info to the child. */
void do_dns_lookup(struct connection *newcon)
{
  int retval;
  static char *buffer = NULL;
  struct dns_child_info *info;

  if (buffer == NULL) {
    buffer = (char *)malloc(sizeof(struct dns_child_info));
    if (buffer == NULL) {
      fprintf(stderr, "do_dns_lookup: malloc() failure.\n");
      return;
    }
  }

  info = (struct dns_child_info *) buffer;

  info->connection_fd   = newcon->fd;
  info->connection_time = newcon->signon;
  memcpy((void *) &(info->connection_addr),
         (void *) &(newcon->connection.sin_addr),
         sizeof(newcon->connection.sin_addr));

  retval = write(dns_fd, buffer, sizeof(struct dns_child_info));
  if (retval < 0) {
    fprintf(stderr, "do_dns_lookup: write to child failed: %s.\n",
            strerror(errno));
  }
}

/* Receive the DNS info from the child. */

void do_dns_received(int fd)
{
  int retval;
  static char *buffer = NULL;
  struct dns_child_info *info;

  struct connection *p;

  if (buffer == NULL) {
    buffer = (char *)malloc(sizeof(struct dns_child_info));
    if (buffer == NULL) {
      fprintf(stderr, "do_dns_lookup: malloc() failure.\n");
      return;
    }
  }

  info = (struct dns_child_info *) buffer;

  retval = read(fd, buffer, sizeof(struct dns_child_info));
  if (retval < 0) {
    fprintf(stderr, "do_dns_received:  Error reading from child: %s.\n",
            strerror(errno));
    return;
  }

  for (p = connections; p != NULL; p = p->next) {
    if (p->fd != info->connection_fd ||
        p->signon != info->connection_time ||
        memcmp((void *) &(p->connection.sin_addr),
               (void *) &(info->connection_addr),
               sizeof(info->connection_addr)) != 0 ||
        p->hostname_resolved != 0)
      /* These aren't the droids we're looking for. */
      continue;

    /* Okay, the fd's, signon times, and addresses agree, and this
       connection's hostname hasn't been previously resolved, so while there
       *may* be some weird potential race conditions here, we've got pretty
       good odds.  And since the addresses agree, we're not actually giving
       something the wrong hostname. */

    if (info->connection_resolved) {
      free(p->hostname);
      p->hostname = strdup(info->connection_hostname);
      p->hostname_resolved = 1;

      if (banned_host(p->hostname)) {
        announce_disconnection(p, strerror(ECONNREFUSED));
        close_connection(p);
      }
    }
    else {
      p->hostname_resolved = -1;
    }
    return;
  }
  fprintf(stderr, "do_dns_received: connection from %s (%s) on fd %d "
          "at time_t %lu not found -- early signoff?\n",
          inet_ntoa(info->connection_addr),
          info->connection_resolved ? info->connection_hostname : "unresolved",
          info->connection_fd, (unsigned long) info->connection_time);
  return;
}

#endif
