/*
 * Dynamic string management.
 *
 * This is used for dealing with "dynamic" strings.  If we fall off the end of
 * the area allocated, reallocate and copy.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 1995 Jonathan Lennox
 * Copyright 2017 Russ Allbery <eagle@eyrie.org>
 *
 * See LICENSE for licensing terms.
 */

#include <config.h>
#include <portable/system.h>

#include <errno.h>

#include "dyn-str.h"

extern char *progname;

/* Append an ordinary string to a dynamic string, re-sizing as necessary. */
/* Return NULL if we've had a malloc() failure, otherwise the dyn_str. */
dyn_string *dyn_strcat(dyn_string *dest, const char *src)
{
  char *new_buf;

  if (dest == NULL || dest->len == 0) return NULL;
  if (dest->buf == NULL) {
    dest->buf = malloc(dest->len);
    if (dest->buf == NULL) {
      dest->len = 0;
      fprintf(stderr, "%s: dyn_strcat: %s\n", progname, strerror(ENOMEM));
      return NULL;
    }
    dest->buf[0] = '\0';
    dest->pos = dest->buf;
  }

  for(; *src != '\0'; dest->pos++, src++) {
    if (dest->pos - dest->buf >= dest->len - 1) {
      /* Buffer is full, double its size. */
      *(dest->pos) = '\0';
      dest->len *= 2;
      new_buf = malloc(dest->len);
      if (new_buf == NULL) {
        dest->buf = NULL;
        dest->len = 0;
        fprintf(stderr, "%s: dyn_strcat: %s\n", progname, strerror(ENOMEM));
        return NULL;
      }
      strcpy(new_buf, dest->buf);
      dest->pos = (dest->pos - dest->buf) + new_buf;
      free(dest->buf);
      dest->buf = new_buf;
      new_buf = NULL;
    }

    *(dest->pos) = *src;
  }

  *(dest->pos) = '\0';

  return dest;
}

/* Analogous -- copy the string to the dynamic string. */
dyn_string *dyn_strcpy(dyn_string *dest, const char *src)
{
  if (dest == NULL) return NULL;

  dest->pos = dest->buf;
  if (dest->pos != NULL) *(dest->pos) = '\0';
  return dyn_strcat(dest, src);
}
