/*
 * Utility functions for the Calvin chatserver.
 *
 * Miscellaneous functions that didn't really fit in anyplace else in the
 * source code.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 1995 Jonathan Lennox
 * Copyright 2017 Russ Allbery <eagle@eyrie.org>
 *
 * See LICENSE for licensing terms.
 */

/* AIX requires this to be the first thing in the file... */
#if !defined (__GNUC__) && defined (_AIX)
 #pragma alloca
#endif

#include <config.h>
#include <portable/system.h>

#include <ctype.h>
#include <errno.h>
#include <regex.h>

#ifdef __GNUC__
# undef alloca
# define alloca(size) __builtin_alloca(size)
#else
# if HAVE_ALLOCA_H
#  include <alloca.h>
# else
#  if !defined(_AIX) && !defined(alloca) /* predefined by HP cc +Olibcalls */
char *alloca ();
#  endif
# endif
#endif

#include <util/macros.h>

#include "channels.h"
#include "connections.h"
#include "format.h"
#include "config-file.h"
#include "nicks.h"
#include "dyn-str.h"
#include "misc.h"

#ifndef RAND_MAX
#define RAND_MAX 0x7fffffff /* XXX is this gonna be right everywhere? */
#endif

extern char *progname;

/* Show the user the current time, as the server understands it. */
int date_cmd(const char *cmd_name UNUSED, struct connection *user,
             char *args UNUSED)
{
  return write_to_user(formatn("current_time", user, NULL),
                       user);
}

/* Repeat the signon message. */
int motd_cmd(const char *cmd_name UNUSED, struct connection *user,
             char *args UNUSED)
{
  return write_to_user(formatn("motd", user, NULL),
                       user);
}

#ifdef DO_ROLL

/* Roll one die */
/* The equation is sort of a mess, but I believe all rounding errors have
   now been eliminated. */
static int roll(int sides)
{
  return (int) ((rand() / (((double) RAND_MAX) + 1)) *
    (double) sides) + 1;
}

struct roll_desc {
  int dice;         /* Number of dice,  0 - (currently) 40. */
  int sides;        /* Number of sides, 0 - (currently) 1000. */
  int half_die;     /* Flag -- do we have a half-die? */
  int champs;       /* Flag -- is this a champions roll? */
  int ka;           /* Flag -- is this a killing attack? */
  int addition;     /* Amount that should be added to/subtracted from the
                       roll. */
  int multiple_bonus; /* Amount that should be added to/subtracted from the
                         stun multiplier for killing attacks. */
};

#define ROLLSBUF_SIZE 1024
#define DESCBUF_SIZE 128
#define MAX_DICE 125
#define MAX_SIDES 9999999
#ifndef _RXH
#define MIN(x,y) ((x)<(y)?(x):(y))
#define MAX(x,y) ((x)>(y)?(x):(y))
#endif

/* Parse a die-roll string, returning in roll */

/* This assumes that the (extended) regexp is generally of the form:
   ^([0-9]*)(\\.5)?[dD]([0-9]*)([+-][0-9]+)([cC]|[kK]([+-][0-9]+))?[ \t]*$
   substr 1 - # of dice.
          2 - optional half-die
          3 - # of sides.
          4 - addition to the base roll.
          5 - if we're a champs or a KA.
          6 - any bonus to stun multiple
   Returns positive on incomprehensible format, negative on error.
*/

static int parse_roll(char *args, struct roll_desc *this_roll, int champs)
{
  regmatch_t substrs[7];
  int retval;
  char *ptr;

  this_roll->dice = -1;
  this_roll->sides = -1;
  this_roll->half_die = 0;
  this_roll->champs = champs;
  this_roll->ka = 0;
  this_roll->addition = 0;
  this_roll->multiple_bonus = 0;

  if (args[0] != '\0') {
    retval = regexec(get_conf_regex(DICE_REGEX), args, 7, substrs, 0);
    if (retval == REG_NOMATCH) {
      return 1; /* XXX is this the right thing to do? */
    }
    else if (retval) {
      fprintf(stderr, "%s: parse_roll: error %d matching string %s\n",
              progname, retval, args);
      return -1;
    } else {
      if (substrs[1].rm_so != substrs[1].rm_eo) {
        this_roll->dice = strtol(args + substrs[1].rm_so, &ptr, 0);
        if (ptr != args + substrs[1].rm_eo)
          fprintf(stderr, "%s: parse_roll: confused on strtol of substr 1\n",
                  progname);
      }
      if (substrs[2].rm_so != substrs[2].rm_eo) {
        this_roll->half_die = 1;
        if (this_roll->dice == -1) this_roll->dice = 0;
      }
      if (substrs[3].rm_so != substrs[3].rm_eo) {
        this_roll->sides = strtol(args + substrs[3].rm_so, &ptr, 0);
        if (ptr != args + substrs[3].rm_eo)
          fprintf(stderr, "%s: parse_roll: confused on strtol of substr 3\n",
                  progname);
      }
      if (substrs[4].rm_so != substrs[4].rm_eo) {
        this_roll->addition = strtol(args + substrs[4].rm_so, &ptr, 0);
        if (ptr != args + substrs[4].rm_eo)
          fprintf(stderr, "%s: parse_roll: confused on strtol of substr 4\n",
                  progname);
      }
      if (substrs[5].rm_so != substrs[5].rm_eo) {
        if (args[substrs[5].rm_so] == 'c' || args[substrs[5].rm_so] == 'C') {
          this_roll->champs = 1;
        }
        else if (args[substrs[5].rm_so] == 'k'
                 || args[substrs[5].rm_so] == 'K') {
          this_roll->champs = 1;
          this_roll->ka = 1;
          if (substrs[6].rm_so != substrs[6].rm_eo) {
            this_roll->multiple_bonus = strtol(args + substrs[6].rm_so,
                                               &ptr, 0);
            if (ptr != args + substrs[6].rm_eo)
              fprintf(stderr, "%s: parse_roll: confused on strtol of substr 6\n",
                      progname);
          }
        }
      }
    }
  }

#if 0
  fprintf(stderr, "\"%s\", %d-%d, %d-%d, %d-%d, %d-%d, %d-%d, %d-%d, %d-%d\n",
          args,
          substrs[0].rm_so, substrs[0].rm_eo,
          substrs[1].rm_so, substrs[1].rm_eo,
          substrs[2].rm_so, substrs[2].rm_eo,
          substrs[3].rm_so, substrs[3].rm_eo,
          substrs[4].rm_so, substrs[4].rm_eo,
          substrs[5].rm_so, substrs[5].rm_eo,
          substrs[6].rm_so, substrs[6].rm_eo);
#endif

  if (this_roll->dice < 0) this_roll->dice = -1;
  if (this_roll->sides < 1) this_roll->sides = -1;

  if (this_roll->dice == -1) {
    if (this_roll->sides != -1)
      this_roll->dice = 1;
    else if (this_roll->ka)
      this_roll->dice = 2; /* XXX Somewhat arbitrary. */
    else if (this_roll->champs)
      this_roll->dice = 6; /* XXX Equally arbitrary. */
    else
      this_roll->dice = 3;
  }
  if (this_roll->sides == -1)
    this_roll->sides = 6;

  /* XXX would it be possible to remove this? */
  if (this_roll->dice > MAX_DICE)
    this_roll->dice = MAX_DICE;
  if (this_roll->sides > MAX_SIDES)
    this_roll->sides = MAX_SIDES;

  return 0;
}

/* Format the result of a die roll. */
/* XXX Is there any way we could get these in chatservrc? */
static void format_roll_result(char *buf, struct roll_desc *this_roll, int
                               stun, int body)
{
  if (this_roll->champs)
    if (this_roll->addition != 0 &&
        this_roll->ka &&
        this_roll->multiple_bonus != 0)
      sprintf(buf, "%d%sd%d%+dk%+d = %d stun, %d body",
              this_roll->dice,
              this_roll->half_die?".5":"",
              this_roll->sides,
              this_roll->addition,
              this_roll->multiple_bonus,
              stun, body);

    else if (this_roll->ka &&
             this_roll->multiple_bonus != 0)
      sprintf(buf, "%d%sd%dk%+d = %d stun, %d body",
              this_roll->dice,
              this_roll->half_die?".5":"",
              this_roll->sides,
              this_roll->multiple_bonus,
              stun, body);

    else if (this_roll->addition)
      sprintf(buf, "%d%sd%d%+d%s = %d stun, %d body",
              this_roll->dice,
              this_roll->half_die?".5":"",
              this_roll->sides,
              this_roll->addition,
              this_roll->ka?"k":"",
              stun, body);

    else
      sprintf(buf, "%d%sd%d%s = %d stun, %d body",
              this_roll->dice,
              this_roll->half_die?".5":"",
              this_roll->sides,
              this_roll->ka?"k":"",
              stun, body);

  else
    if (this_roll->addition != 0)
      sprintf(buf, "%d%sd%d%+d = %d",
              this_roll->dice,
              this_roll->half_die?".5":"",
              this_roll->sides,
              this_roll->addition,
              stun);

    else
      sprintf(buf, "%d%sd%d = %d",
              this_roll->dice,
              this_roll->half_die?".5":"",
              this_roll->sides,
              stun);

  return;
}

/* Actually roll the dice. */
static int roll_dice(char *args, char *rollsbuf, char *descbuf, int champs)
{
  int this_die, stun = 0, body = 0, multiple;
  int i, retval, flag;
  char *ptr;
  struct roll_desc this_roll;

  retval = parse_roll(args, &this_roll, champs);
  if (retval) return retval;

  ptr = rollsbuf;
  flag = 0;
  *ptr = '\0';

  for (i = 0; i < this_roll.dice; i++) {
    this_die = roll(this_roll.sides);
    if (this_roll.champs && this_roll.ka)
      body += this_die;
    else
      stun += this_die;
    if (this_roll.champs && !this_roll.ka)
      body += (this_die == this_roll.sides) ? 2 : (this_die == 1) ? 0 : 1;
    sprintf(ptr, "%s%d", flag?" ":"", this_die);
    flag = 1;
    ptr = strchr(ptr, '\0');
  }

  /* XXX Half-dice aren't terribly well tested for dice with an odd number
     of sides. */
  if (this_roll.half_die) {
    this_die = roll(this_roll.sides);
    if (this_roll.champs && this_roll.ka)
      body += (this_die + 1)/2;
    else
      stun += (this_die + 1)/2;
    if (this_roll.champs && !this_roll.ka)
      body += (this_die - 1) / ((this_roll.sides + 1)/ 2); /* XXX ?? */
    sprintf(ptr, "%s[%d]", flag?" ":"", this_die);
    flag = 1;
    ptr=strchr(ptr, '\0');
  }

  if (this_roll.champs && this_roll.ka) {
    body += this_roll.addition;
    multiple = roll(this_roll.sides) + this_roll.multiple_bonus - 1;
    multiple = MAX(multiple, 1);
    sprintf(ptr, " x%d", multiple);
    ptr = strchr(ptr, '\0');
    stun = body * multiple;
  } else {
    stun += this_roll.addition;
  }

  format_roll_result(descbuf, &this_roll, stun, body);

  return 0;
}

/* For forging the roll -- grab the first int off the ptr, setting *p to the
   point after that ptr, or NULL if none found.  Make sure the roll is
   valid. Returns -1 if anything's gone wrong.*/
static int pick_off_roll(char **p, int min, int max)
{
  int r;
  char *q;

  if (p == NULL || *p == NULL) return -1;

  while (isspace(**p))
    (*p)++;

  r = strtol(*p, &q, 0);
  if (*p == q || r < min || r > max) {
    *p = NULL;
    return -1;
  }

  *p = q;

  while (isspace(**p))
    (*p)++;

  return r;
}

/* Pretend to roll the dice. */
/* XXX: formats are a little ucky. */
static int forge_roll(char *args, char *rollsbuf, char *descbuf, int champs)
{
  int this_die, stun = 0, body = 0, multiple;
  int i, retval, flag;
  char *ptr, *rolls, *p;
  struct roll_desc this_roll;

  rolls = p = strchr(args, '=');
  if (rolls == NULL) return -1;

  *rolls = '\0';

  retval = parse_roll(args, &this_roll, champs);
  *rolls = '=';
  if (retval) return retval;

  rolls++;

  ptr = rollsbuf;
  flag = 0;
  *ptr = '\0';

  for (i = 0; i < this_roll.dice; i++) {
    this_die = pick_off_roll(&rolls, 1, this_roll.sides);
    if (this_die < 0) return this_die;
    if (this_roll.champs && this_roll.ka)
      body += this_die;
    else
      stun += this_die;
    if (this_roll.champs && !this_roll.ka)
      body += (this_die == this_roll.sides) ? 2 : (this_die == 1) ? 0 : 1;
    sprintf(ptr, "%s%d", flag?" ":"", this_die);
    flag = 1;
    ptr = strchr(ptr, '\0');
  }

  /* XXX Half-dice aren't terribly well tested for dice with an odd number
     of sides. */
  if (this_roll.half_die) {
    if (*rolls == '[')
      rolls++;
    this_die = pick_off_roll(&rolls, 1, this_roll.sides);
    if (this_die < 0) return this_die;
    if (*rolls == ']')
      do rolls++; while (isspace(*rolls));

    if (this_roll.champs && this_roll.ka)
      body += ((unsigned int) this_die + 1)/2;
    else
      stun += ((unsigned int) this_die + 1)/2;
    if (this_roll.champs && !this_roll.ka)
      body += (this_die - 1) / ((this_roll.sides + 1)/ 2); /* XXX ?? */

    sprintf(ptr, "%s[%d]", flag?" ":"", this_die);
    flag = 1;
    ptr=strchr(ptr, '\0');
  }

  if (this_roll.champs && this_roll.ka) {
    body += this_roll.addition;
    if (*rolls == 'x') {
      rolls++;
      multiple = pick_off_roll(&rolls, MAX(this_roll.multiple_bonus, 1),
                               this_roll.sides + this_roll.multiple_bonus - 1);
      if (multiple < 0) return multiple;
    } else {
      multiple = pick_off_roll(&rolls, 1, this_roll.sides);
      if (multiple < 0) return multiple;
      multiple += this_roll.multiple_bonus - 1;
    }
    multiple = MAX(multiple, 1);
    sprintf(ptr, " x%d", multiple);
    ptr = strchr(ptr, '\0');
    stun = body * multiple;
  } else {
    stun += this_roll.addition;
  }

  format_roll_result(descbuf, &this_roll, stun, body);

  if (*rolls != '\0') return -1;

  return 0;
}

/* Roll some dice. */
int roll_cmd(const char *cmd_name UNUSED, struct connection *user, char *args)
{
  char rollsbuf[ROLLSBUF_SIZE], descbuf[DESCBUF_SIZE];
  int retval;
  char *roll_str, *roll_str_dup;

  retval = roll_dice(args, rollsbuf, descbuf, (cmd_name[0] == 'C'));

  if (retval) {
    retval = write_to_user(formatn("bad_dice_arg", user, args),
                           user);
    return 0;
  }

/* TODO: Fix this. */
  roll_str = format2(lookup_conf("roll_format"), user, descbuf, rollsbuf);
  roll_str_dup = alloca(strlen(roll_str) + 1);
  if (roll_str_dup == NULL) return -ENOMEM;
  strcpy(roll_str_dup, roll_str);

  return send_to_recipient(user, roll_str_dup, "roll", user->target);
}

/* Pretend to roll some dice. Available at priv level 1. */
int forgeroll_cmd(const char *cmd_name UNUSED, struct connection *user,
                  char *args)
{
  char rollsbuf[ROLLSBUF_SIZE], descbuf[DESCBUF_SIZE];
  int retval;
  char *roll_str, *roll_str_dup;

  retval = forge_roll(args, rollsbuf, descbuf, (cmd_name[0] == 'C'));

  if (retval) {
    retval = write_to_user(formatn("bad_forgery", user, args),
                           user);
    return 0;
  }

/* TODO: Fix this. */
  roll_str = format2(lookup_conf("roll_format"), user, descbuf, rollsbuf);
  roll_str_dup = alloca(strlen(roll_str) + 1);
  if (roll_str_dup == NULL) return -ENOMEM;
  strcpy(roll_str_dup, roll_str);

  return send_to_recipient(user, roll_str_dup, "roll", user->target);
}

#endif
