%{
/*
 * This is the lex parser for the configuration file.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 1995 Jonathan Lennox
 * Copyright 2017 Russ Allbery <eagle@eyrie.org>
 *
 * See LICENSE for licensing terms.
 */

#include <config.h>
#include <portable/system.h>

#include <errno.h>

#include <util/macros.h>

#include "scan-config.h"
/* #include "config-file.h" */
#include "dyn-str.h"

/* Disable some warnings that the flex-generated parser does not pass. */
#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ > 2) || defined(__clang__)
# pragma GCC diagnostic ignored "-Wsign-compare"
# pragma GCC diagnostic ignored "-Wunused-function"
#endif

#define YY_DECL static int do_scan(struct configuration *conf)

static int lineno = 1;
static dyn_string name_str = DYN_STR_INIT(64),   *name = &name_str;
static dyn_string val_str  = DYN_STR_INIT(1024), *val = &val_str;
static int retval;

#define ADD_TEXT(buf) do { \
                        val = dyn_strcat(val, buf); \
                        if (val == NULL) return(-ENOMEM); \
                       } while (0)

#define ADD_YYTEXT ADD_TEXT(yytext)

%}
%s value
%x value_string
%%

<INITIAL>[[:alnum:]_-]+   { /* Read a name */
                            name = dyn_strcpy(name, yytext);
                            if (name == NULL) return(-ENOMEM);
                            val = dyn_strcpy(val, "");
                            if (name == NULL) return(-ENOMEM);
                            BEGIN(value);
                          }

<value>\"                 BEGIN(value_string);
<value_string>{
  \"     {  /* End of string */
            retval = set_conf_in_config(name->buf, val->buf, conf);
            dyn_strcpy(name, "");
            if (name == NULL) return (-ENOMEM);
            BEGIN(INITIAL);
          }

  \n      { lineno++; ADD_YYTEXT; }

  \\[0-7]{1,3} {
          /* Octal escape sequence */
          unsigned int result;

          sscanf( yytext +1, "%o", &result );
          if ( result > 0xff )
            return 1;
          yytext[0] = result;
          yytext[1] = '\0';
          ADD_YYTEXT;
        }

  \\[0-9]+ {
          /* Bad escape sequence */
          return(1);
        }

  /* XXX I don't want to be using yyleng here, or elsewhere. */
  \\n  { ADD_TEXT("\n"); }
  \\t  { ADD_TEXT("\t"); }
  \\v  { ADD_TEXT("\v"); }
  \\b  { ADD_TEXT("\b"); }
  \\r  { ADD_TEXT("\r"); }
  \\f  { ADD_TEXT("\f"); }
  \\a  { ADD_TEXT("\a"); }

  \\[^\n]  { ADD_TEXT(yytext + 1); }
  \\\n     { lineno++; ADD_TEXT(yytext + 1); }

  [^\\\n\"]+ { ADD_YYTEXT; }
}

#[^\n]+                   /* Eat comments */
[ \t]+                    /* Eat whitespace */
"\n"                      lineno++;
<INITIAL><<EOF>>          yyterminate();
<value,value_string><<EOF>>  return 1;
.                         return 1;

%%
int scan_config(FILE *config_file, struct configuration *conf)
{
  int error;

  yyin = config_file;

  error = do_scan(conf);
  if (error < 0)
    return error;
  else if (error > 0)
    return lineno;
  else return 0;
}

int yywrap(void) { return 1; /* Always done. */ }
