/*
 * TLS connection handling.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 2005 Matthew Gerber
 * Copyright 2017 Russ Allbery <eagle@eyrie.org>
 *
 * See LICENSE for licensing terms.
 */

#include <config.h>
#include <portable/system.h>

#include <sys/stat.h>
#include <time.h>

#include "channels.h"
#include "connections.h"
#include "config-file.h"

#ifdef GNUTLS11
static gnutls_anon_server_credentials_t credentials;
static gnutls_certificate_credentials_t xcred = NULL;
static time_t credential_time;

static gnutls_dh_params_t dh_params;

static void ssl_generate_dh_params(void) {
  gnutls_dh_params_init(&dh_params);
  gnutls_dh_params_generate2(dh_params, 1024);
}

static void
load_credentials(void) {
  const char *certfile = lookup_conf("ssl_cert");
  const char *keyfile = lookup_conf("ssl_key");
  struct stat st;

  if (stat(certfile, &st) < 0) {
    if (xcred == NULL) {
      fprintf(stderr, "Could not allocate server credentials.\n");
      exit(1);
    } else {
      fprintf(stderr,
              "Cannot stat certificate file, using cached certificate\n");
      return;
    }
  }
  if (xcred != NULL)
    gnutls_certificate_free_credentials(xcred);
  if (gnutls_certificate_allocate_credentials(&xcred) != GNUTLS_E_SUCCESS) {
    fprintf(stderr, "Could not allocate server credentials.\n");
    exit(1);
  }
  gnutls_certificate_set_x509_key_file(xcred, certfile, keyfile,
                                       GNUTLS_X509_FMT_PEM);
  gnutls_certificate_set_dh_params(xcred, dh_params);
  credential_time = st.st_ctime;
  printf("Loaded certificate from %s (key: %s)\n", certfile, keyfile);
}

void ssl_init(void) {
  if(gnutls_global_init()) {
    fprintf(stderr, "Error initializing gnutls.\n");
    exit(1);
  }

  ssl_generate_dh_params();
  if(gnutls_anon_allocate_server_credentials(&credentials)) {
    fprintf(stderr, "Could not allocate server credentials.\n");
    exit(1);
  }
  gnutls_anon_set_server_dh_params(credentials, dh_params);

  load_credentials();
  printf("Secure Sockets Layer initialization complete.\n");
}

void ssl_end(SSL ssl) {
  gnutls_bye(ssl, GNUTLS_SHUT_RDWR);
  gnutls_deinit(ssl);
}

int ssl_write(SSL ssl, char *s, int len) {
  return gnutls_write(ssl, s, len);
}

int ssl_read(SSL ssl, char *s, int len) {
  return gnutls_read(ssl, s, len);
}

struct connection *start_ssl_setup(struct connection *newcon, int fd) {
  gnutls_session_t session;
  struct stat st;
  const char *certfile = lookup_conf("ssl_cert");

  if (gnutls_init(&session, GNUTLS_SERVER) != GNUTLS_E_SUCCESS) {
    close_connection(newcon);
    return NULL;
  }

  if (stat(certfile, &st) < 0)
    fprintf(stderr, "Cannot stat certificate file, using cached certificate\n");
  else if (st.st_ctime != credential_time)
    load_credentials();

  gnutls_set_default_priority(session);
  gnutls_session_enable_compatibility_mode(session);
  gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, xcred);
  gnutls_credentials_set(session, GNUTLS_CRD_ANON, credentials);
  gnutls_dh_set_prime_bits(session, 1024);

  gnutls_transport_set_int(session, fd);

  newcon->ssl_finished = 0;
  newcon->ssl = session;
  continue_ssl_setup(newcon);
  return newcon;
}

void continue_ssl_setup(struct connection *current) {
  int retval;

  retval = gnutls_handshake(current->ssl);
  if(retval == 0) {
    current->ssl_finished = 1;
  } else {
    if(retval == GNUTLS_E_AGAIN || retval == GNUTLS_E_INTERRUPTED) {
      current->ssl_finished = 0;
    } else {
      fprintf(stderr, "SSL FATAL: ");
      gnutls_perror(retval);
      close_connection(current);
    }
  }
}
#endif

#ifndef DO_SSL
void ssl_init() {}
void ssl_end(SSL ssl) {}
int ssl_write(SSL ssl, char *s, int len) {}
int ssl_read(SSL ssl, char *s, int len) {}
struct connection *start_ssl_setup(struct connection *newcon, int fd) {}
void continue_ssl_setup(struct connection *current) {}
#endif
