/*
 * Functions for fairly rudimentary handling of telnet proto.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 1995 Jonathan Lennox
 * Copyright 2017 Russ Allbery <eagle@eyrie.org>
 *
 * See LICENSE for licensing terms.
 */

/* AIX requires this to be the first thing in the file... */
#if !defined (__GNUC__) && defined (_AIX)
 #pragma alloca
#endif

#include <config.h>
#include <portable/system.h>

#include <errno.h>

#ifdef __GNUC__
# undef alloca
# define alloca(size) __builtin_alloca(size)
#else
# if HAVE_ALLOCA_H
#  include <alloca.h>
# else
#  if !defined(_AIX) && !defined(alloca) /* predefined by HP cc +Olibcalls */
char *alloca ();
#  endif
# endif
#endif

#include "channels.h"
#include "connections.h"
#include "telnet.h"

/* Telnet things */
#define IAC  255

#define WILL 251
#define WONT 252
#define DO   253
#define DONT 254

#define NOP  241
#define AYT  246
#define EC   247
#define EL   248

extern char *progname;

/* Read from a FD, understanding telnet options.
   Return value: -2, error.
                 -1, EOF.
                 0, no data.
                 positive num, data bytes returned.
   */
int telnet_read(struct connection *user, char *buf, int buf_len)
{
  int retval, i, j, escaped_chars = 0;
  unsigned char *local_buf;
  int retval2;
  char little_buf[4];

  int fd = user->fd;
  SSL ssl = user->ssl;

  if (buf_len <= 0) return 0;

  local_buf = alloca(buf_len);
  if (local_buf == NULL) {
    /* Fake it. */
    fprintf(stderr, "%s: telnet_read: connection %d: "
            "Out of memory, faking it.\n",
            progname, fd);
#ifdef DO_SSL
    return ssl ? ssl_read(ssl, buf, buf_len) : read(fd, buf, buf_len);
#else
    return read(fd, buf, buf_len);
#endif
  }

#ifdef DO_SSL
  retval=ssl ? ssl_read(ssl, (char *) local_buf, buf_len) :
         read(fd, local_buf, buf_len);
#else
  retval=read(fd, local_buf, buf_len);
#endif

  if (retval <= 0) {
    return retval - 1;
  }

  for (i = 0, j = 0; i < retval; i++) {
    if ((local_buf[i] >= 7 && local_buf[i] <= 13) ||
        (local_buf[i] >= 32 && local_buf[i] <= 126) ||
        (local_buf[i] >= 160 && local_buf[i] <= 254))
      buf[j++] = local_buf[i];
    else if (local_buf[i] == 127)
      /* Translate ^? into ^H for backspace_cleanup to take care of */
      buf[j++] = 'H' - 'A' + 1;
    else if (local_buf[i] >= 135 && local_buf[i] <= 141)
      buf[j++] = local_buf[i] & 0177;
    else if (local_buf[i] == IAC) {
      /* XXX I'm assuming here that telnet options all come through in
         one read. */
      switch (local_buf[++i]) {
      case DO:
      case WILL:
        sprintf(little_buf, "%c%c%c", IAC, (local_buf[i]==DO ? WONT : DONT),
                             local_buf[i+1]);
        /* Note that we assume here that the fd is a socket. */
        retval2 = queue_for_write(user, little_buf, 3);
        if (retval2 != 3) {
#if NOISY_TELNET_HANDLING
          fprintf(stderr, "%s: telnet_read: connection %d: "
                  "trouble writing %s %u: %s\n",
                  progname, fd, (local_buf[i]==DO ? "WONT" : "DONT"),
                  (unsigned char) local_buf[i+1],
                  strerror(errno));
#endif
        } else {
#if NOISY_TELNET_HANDLING
          fprintf(stderr, "%s: telnet_read: connection %d: "
                  "remote telnet sent IAC %s %u, responded %s\n",
                  progname, fd, (local_buf[i]==DO ? "DO" : "WILL"),
                  (unsigned char) local_buf[i+1],
                  (local_buf[i]==DO ? "WONT" : "DONT"));
#endif
        }
        i++;
        escaped_chars += 3;
        break;
      case DONT:
      case WONT:
#if NOISY_TELNET_HANDLING
        fprintf(stderr, "%s: telnet_read: connection %d: "
                "remote telnet sent IAC %s %u, ignoring\n",
                progname, fd, (local_buf[i]==DONT ? "DONT" : "WONT"),
                (unsigned char) local_buf[i+1]);
#endif
        i++;
        escaped_chars += 3;
        break;
      case AYT:
        retval2 = queue_for_write(user, "\r\n[Yes]\r\n", 9);
        if (retval2 != 9) {
#if NOISY_TELNET_HANDLING
          fprintf(stderr, "%s: telnet_read: connection %d: "
                  "trouble writing \"\\r\\n[Yes]\\r\\n\": %s\n",
                  progname, fd, strerror(errno));
#endif
        } else {
#if NOISY_TELNET_HANDLING
          fprintf(stderr, "%s: telnet_read: connection %d: "
                  "remote telnet sent IAC AYT, responded "
                  "\"\\r\\n[Yes]\\r\\n\"\n",
                  progname, fd);
#endif
        }
        break;
      case EC:
#if NOISY_TELNET_HANDLING
        fprintf(stderr, "%s: telnet_read: connection %d: "
                "remote telnet sent IAC EC, inserting ^H\n",
                progname, fd);
#endif
        buf[j++] = 'H' - 'A' + 1;
        escaped_chars += 1;
        break;
      case EL:
#if NOISY_TELNET_HANDLING
        fprintf(stderr, "%s: telnet_read: connection %d: "
                "remote telnet sent IAC EL, inserting ^U\n",
                progname, fd);
#endif
        buf[j++] = 'U' - 'A' + 1;
        escaped_chars += 1;
        break;
      case IAC:
#if NOISY_TELNET_HANDLING
        fprintf(stderr, "%s: telnet_read: connection %d: "
                "remote telnet sent IAC IAC, quoting\n",
                progname, fd);
#endif
        buf[j++] = IAC;
        escaped_chars += 1;
        break;
      default:
#if NOISY_TELNET_HANDLING
        fprintf(stderr, "%s: telnet_read: connection %d: "
                "remote telnet sent IAC %u, ignoring\n",
                progname, fd,
                (unsigned char) local_buf[i]);
#endif
        escaped_chars += 2;
      }
    } else {
      /* Drop all other characters on the floor. */
      escaped_chars += 1;
    }
  }

#if NOISY_TELNET_HANDLING
  if (j + escaped_chars != i)
    fprintf(stderr, "%s: telnet_read: connection %d: "
            "wuh-oh, (j == %d) + (escaped_chars == %d) != (i == %d)\n",
            progname, fd, j, escaped_chars, i);
#endif

  return i - escaped_chars;
}
